@@ -13,37 +13,47 @@
 #include "x86-linux-ptrace.h"
 #include "x86-arch.h"
 
+typedef struct
+{
+	int slot;
+	int insn_size;
+	gboolean update_ip;
+	guint32 code_address;
+	guint32 original_eip;
+} CodeBufferData;
+
 struct ArchInfo
 {
-	guint32 call_address;
-	guint64 callback_argument;
 	INFERIOR_REGS_TYPE current_regs;
 	INFERIOR_FPREGS_TYPE current_fpregs;
-	INFERIOR_REGS_TYPE *saved_regs;
-	INFERIOR_FPREGS_TYPE *saved_fpregs;
-	int saved_signal;
-	GPtrArray *rti_stack;
+	GPtrArray *callback_stack;
+	CodeBufferData *code_buffer;
 	guint64 dr_control, dr_status;
 	int dr_regs [DR_NADDR];
 };
 
 typedef struct
 {
-	INFERIOR_REGS_TYPE *saved_regs;
-	INFERIOR_FPREGS_TYPE *saved_fpregs;
-	int saved_signal;
-	guint32 call_address;
-	guint32 exc_address;
-	gboolean debug;
+	INFERIOR_REGS_TYPE saved_regs;
+	INFERIOR_FPREGS_TYPE saved_fpregs;
 	guint64 callback_argument;
-} RuntimeInvokeData;
+	guint32 call_address;
+	guint32 stack_pointer;
+	guint32 rti_frame;
+	guint32 exc_address;
+	guint32 pushed_registers;
+	guint32 data_pointer;
+	guint32 data_size;
+	int saved_signal;
+	gboolean debug;
+} CallbackData;
 
 ArchInfo *
 x86_arch_initialize (void)
 {
 	ArchInfo *arch = g_new0 (ArchInfo, 1);
 
-	arch->rti_stack = g_ptr_array_new ();
+	arch->callback_stack = g_ptr_array_new ();
 
 	return arch;
 }
@@ -51,9 +61,7 @@ x86_arch_initialize (void)
 void
 x86_arch_finalize (ArchInfo *arch)
 {
-	g_ptr_array_free (arch->rti_stack, TRUE);
-	g_free (arch->saved_regs);
-     	g_free (arch->saved_fpregs);
+	g_ptr_array_free (arch->callback_stack, TRUE);
 	g_free (arch);
 }
 
@@ -125,6 +133,7 @@ server_ptrace_call_method (ServerHandle *handle, guint64 method_address,
 	ServerCommandError result = COMMAND_ERROR_NONE;
 	ArchInfo *arch = handle->arch;
 	guint32 new_esp, call_disp;
+	CallbackData *cdata;
 
 	guint8 code[] = { 0x68, 0x00, 0x00, 0x00, 0x00, 0x68, 0x00, 0x00,
 			  0x00, 0x00, 0x68, 0x00, 0x00, 0x00, 0x00, 0x68,
@@ -132,16 +141,16 @@ server_ptrace_call_method (ServerHandle *handle, guint64 method_address,
 			  0x00, 0xcc };
 	int size = sizeof (code);
 
-	if (arch->saved_regs)
-		return COMMAND_ERROR_RECURSIVE_CALL;
+	cdata = g_new0 (CallbackData, 1);
 
 	new_esp = (guint32) INFERIOR_REG_ESP (arch->current_regs) - size;
 
-	arch->saved_regs = g_memdup (&arch->current_regs, sizeof (arch->current_regs));
-	arch->saved_fpregs = g_memdup (&arch->current_fpregs, sizeof (arch->current_fpregs));
-	arch->call_address = new_esp + 26;
-	arch->callback_argument = callback_argument;
-	arch->saved_signal = handle->inferior->last_signal;
+	memcpy (&cdata->saved_regs, &arch->current_regs, sizeof (arch->current_regs));
+	memcpy (&cdata->saved_fpregs, &arch->current_fpregs, sizeof (arch->current_fpregs));
+	cdata->call_address = new_esp + 26;
+	cdata->stack_pointer = new_esp - 16;
+	cdata->callback_argument = callback_argument;
+	cdata->saved_signal = handle->inferior->last_signal;
 	handle->inferior->last_signal = 0;
 
 	call_disp = (int) method_address - new_esp;
@@ -158,6 +167,8 @@ server_ptrace_call_method (ServerHandle *handle, guint64 method_address,
 
 	INFERIOR_REG_ESP (arch->current_regs) = INFERIOR_REG_EIP (arch->current_regs) = new_esp;
 
+	g_ptr_array_add (arch->callback_stack, cdata);
+
 	result = _server_ptrace_set_registers (handle->inferior, &arch->current_regs);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
@@ -172,40 +183,49 @@ server_ptrace_call_method (ServerHandle *handle, guint64 method_address,
 
 static ServerCommandError
 server_ptrace_call_method_1 (ServerHandle *handle, guint64 method_address,
-			     guint64 method_argument, const gchar *string_argument,
+			     guint64 method_argument, guint64 data_argument,
+			     guint64 data_argument2, const gchar *string_argument,
 			     guint64 callback_argument)
 {
 	ServerCommandError result = COMMAND_ERROR_NONE;
 	ArchInfo *arch = handle->arch;
 	guint32 new_esp, call_disp;
+	CallbackData *cdata;
 
 	static guint8 static_code[] = { 0x68, 0x00, 0x00, 0x00, 0x00, 0x68, 0x00, 0x00,
-					0x00, 0x00, 0x68, 0x00, 0x00, 0x00, 0x00, 0xe8,
-					0x00, 0x00, 0x00, 0x00, 0xcc };
+					0x00, 0x00, 0x68, 0x00, 0x00, 0x00, 0x00, 0x68,
+					0x00, 0x00, 0x00, 0x00, 0x68, 0x00, 0x00, 0x00,
+					0x00, 0x68, 0x00, 0x00, 0x00, 0x00, 0x68, 0x00,
+					0x00, 0x00, 0x00, 0xe8, 0x00, 0x00, 0x00, 0x00,
+					0xcc };
 	int static_size = sizeof (static_code);
 	int size = static_size + strlen (string_argument) + 1;
 	guint8 *code = g_malloc0 (size);
 	memcpy (code, static_code, static_size);
 	strcpy ((char *) (code + static_size), string_argument);
 
-	if (arch->saved_regs)
-		return COMMAND_ERROR_RECURSIVE_CALL;
+	cdata = g_new0 (CallbackData, 1);
 
 	new_esp = (guint32) INFERIOR_REG_ESP (arch->current_regs) - size;
 
-	arch->saved_regs = g_memdup (&arch->current_regs, sizeof (arch->current_regs));
-	arch->saved_fpregs = g_memdup (&arch->current_fpregs, sizeof (arch->current_fpregs));
-	arch->call_address = new_esp + 21;
-	arch->callback_argument = callback_argument;
-	arch->saved_signal = handle->inferior->last_signal;
+	memcpy (&cdata->saved_regs, &arch->current_regs, sizeof (arch->current_regs));
+	memcpy (&cdata->saved_fpregs, &arch->current_fpregs, sizeof (arch->current_fpregs));
+	cdata->call_address = new_esp + static_size;
+	cdata->stack_pointer = new_esp - 28;
+	cdata->callback_argument = callback_argument;
+	cdata->saved_signal = handle->inferior->last_signal;
 	handle->inferior->last_signal = 0;
 
 	call_disp = (int) method_address - new_esp;
 
-	*((guint32 *) (code+1)) = new_esp + 21;
-	*((guint32 *) (code+6)) = method_argument >> 32;
-	*((guint32 *) (code+11)) = method_argument & 0xffffffff;
-	*((guint32 *) (code+16)) = call_disp - 20;
+	*((guint32 *) (code+1)) = new_esp + static_size;
+	*((guint32 *) (code+6)) = data_argument2 >> 32;
+	*((guint32 *) (code+11)) = data_argument2 & 0xffffffff;
+	*((guint32 *) (code+16)) = data_argument >> 32;
+	*((guint32 *) (code+21)) = data_argument & 0xffffffff;
+	*((guint32 *) (code+26)) = method_argument >> 32;
+	*((guint32 *) (code+31)) = method_argument & 0xffffffff;
+	*((guint32 *) (code+36)) = call_disp - 40;
 
 	result = server_ptrace_write_memory (handle, (guint32) new_esp, size, code);
 	if (result != COMMAND_ERROR_NONE)
@@ -213,6 +233,8 @@ server_ptrace_call_method_1 (ServerHandle *handle, guint64 method_address,
 
 	INFERIOR_REG_ESP (arch->current_regs) = INFERIOR_REG_EIP (arch->current_regs) = new_esp;
 
+	g_ptr_array_add (arch->callback_stack, cdata);
+
 	result = _server_ptrace_set_registers (handle->inferior, &arch->current_regs);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
@@ -222,24 +244,22 @@ server_ptrace_call_method_1 (ServerHandle *handle, guint64 method_address,
 
 static ServerCommandError
 server_ptrace_call_method_2 (ServerHandle *handle, guint64 method_address,
-			     guint64 method_argument, guint64 callback_argument)
+			     guint32 data_size, gconstpointer data_buffer,
+			     guint64 callback_argument)
 {
 	ServerCommandError result = COMMAND_ERROR_NONE;
 	ArchInfo *arch = handle->arch;
-	RuntimeInvokeData *rdata;
+	CallbackData *cdata;
 	guint32 new_esp;
 
-	int size = 57;
+	int size = 57 + data_size;
 	guint8 *code = g_malloc0 (size);
 
-	if (arch->saved_regs)
-		return COMMAND_ERROR_RECURSIVE_CALL;
-
 	new_esp = INFERIOR_REG_ESP (arch->current_regs) - size;
 
 	*((guint32 *) code) = new_esp + size - 1;
 	*((guint64 *) (code+4)) = new_esp + 20;
-	*((guint64 *) (code+12)) = method_argument;
+	*((guint64 *) (code+12)) = new_esp + 56;
 	*((guint32 *) (code+20)) = INFERIOR_REG_EAX (arch->current_regs);
 	*((guint32 *) (code+24)) = INFERIOR_REG_EBX (arch->current_regs);
 	*((guint32 *) (code+28)) = INFERIOR_REG_ECX (arch->current_regs);
@@ -249,17 +269,25 @@ server_ptrace_call_method_2 (ServerHandle *handle, guint64 method_address,
 	*((guint32 *) (code+44)) = INFERIOR_REG_ESI (arch->current_regs);
 	*((guint32 *) (code+48)) = INFERIOR_REG_EDI (arch->current_regs);
 	*((guint32 *) (code+52)) = INFERIOR_REG_EIP (arch->current_regs);
-	*((guint8 *) (code+56)) = 0xcc;
+	*((guint8 *) (code+data_size+56)) = 0xcc;
 
-	rdata = g_new0 (RuntimeInvokeData, 1);
-	rdata->saved_regs = g_memdup (&arch->current_regs, sizeof (arch->current_regs));
-	rdata->saved_fpregs = g_memdup (&arch->current_fpregs, sizeof (arch->current_fpregs));
-	rdata->call_address = new_esp + size;
-	rdata->exc_address = 0;
-	rdata->callback_argument = callback_argument;
-	rdata->saved_signal = handle->inferior->last_signal;
+	cdata = g_new0 (CallbackData, 1);
+	memcpy (&cdata->saved_regs, &arch->current_regs, sizeof (arch->current_regs));
+	memcpy (&cdata->saved_fpregs, &arch->current_fpregs, sizeof (arch->current_fpregs));
+	cdata->call_address = new_esp + size;
+	cdata->stack_pointer = new_esp + 4;
+	cdata->exc_address = 0;
+	cdata->callback_argument = callback_argument;
+	cdata->saved_signal = handle->inferior->last_signal;
+	cdata->pushed_registers = new_esp + 4;
 	handle->inferior->last_signal = 0;
 
+	if (data_size > 0) {
+		memcpy (code+56, data_buffer, data_size);
+		cdata->data_pointer = new_esp + 56;
+		cdata->data_size = data_size;
+	}
+
 	server_ptrace_write_memory (handle, (guint32) new_esp, size, code);
 	g_free (code);
 	if (result != COMMAND_ERROR_NONE)
@@ -268,7 +296,7 @@ server_ptrace_call_method_2 (ServerHandle *handle, guint64 method_address,
 	INFERIOR_REG_EIP (arch->current_regs) = method_address;
 	INFERIOR_REG_ESP (arch->current_regs) = new_esp;
 
-	g_ptr_array_add (arch->rti_stack, rdata);
+	g_ptr_array_add (arch->callback_stack, cdata);
 
 	result = _server_ptrace_set_registers (handle->inferior, &arch->current_regs);
 	if (result != COMMAND_ERROR_NONE)
@@ -286,7 +314,7 @@ server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 {
 	ServerCommandError result = COMMAND_ERROR_NONE;
 	ArchInfo *arch = handle->arch;
-	RuntimeInvokeData *rdata;
+	CallbackData *cdata;
 	guint32 new_esp;
 	int i;
 
@@ -303,9 +331,6 @@ server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 	memcpy (code, static_code, static_size);
 	memcpy (code + static_size, blob_data, blob_size);
 
-	if (arch->saved_regs)
-		return COMMAND_ERROR_RECURSIVE_CALL;
-
 	new_esp = (guint32) INFERIOR_REG_ESP (arch->current_regs) - size;
 	blob_start = new_esp + static_size;
 
@@ -322,14 +347,15 @@ server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 	*((guint32 *) (code+12)) = new_esp + static_size + blob_size + 4;
 	*((guint32 *) (code+16)) = new_esp + 20;
 
-	rdata = g_new0 (RuntimeInvokeData, 1);
-	rdata->saved_regs = g_memdup (&arch->current_regs, sizeof (arch->current_regs));
-	rdata->saved_fpregs = g_memdup (&arch->current_fpregs, sizeof (arch->current_fpregs));
-	rdata->call_address = new_esp + static_size;
-	rdata->exc_address = new_esp + 20;
-	rdata->callback_argument = callback_argument;
-	rdata->debug = debug;
-	rdata->saved_signal = handle->inferior->last_signal;
+	cdata = g_new0 (CallbackData, 1);
+	memcpy (&cdata->saved_regs, &arch->current_regs, sizeof (arch->current_regs));
+	memcpy (&cdata->saved_fpregs, &arch->current_fpregs, sizeof (arch->current_fpregs));
+	cdata->call_address = new_esp + static_size;
+	cdata->stack_pointer = new_esp + 4;
+	cdata->exc_address = new_esp + 20;
+	cdata->callback_argument = callback_argument;
+	cdata->debug = debug;
+	cdata->saved_signal = handle->inferior->last_signal;
 	handle->inferior->last_signal = 0;
 
 	result = server_ptrace_write_memory (handle, (guint32) new_esp, size, code);
@@ -339,8 +365,9 @@ server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 
 	INFERIOR_REG_EIP (arch->current_regs) = invoke_method;
 	INFERIOR_REG_ESP (arch->current_regs) = new_esp;
+	INFERIOR_REG_EBP (arch->current_regs) = 0;
 
-	g_ptr_array_add (arch->rti_stack, rdata);
+	g_ptr_array_add (arch->callback_stack, cdata);
 
 	result = _server_ptrace_set_registers (handle->inferior, &arch->current_regs);
 	if (result != COMMAND_ERROR_NONE)
@@ -366,13 +393,13 @@ check_breakpoint (ServerHandle *handle, guint64 address, guint64 *retval)
 	return TRUE;
 }
 
-static RuntimeInvokeData *
-get_runtime_invoke_data (ArchInfo *arch)
+static CallbackData *
+get_callback_data (ArchInfo *arch)
 {
-	if (!arch->rti_stack->len)
+	if (!arch->callback_stack->len)
 		return NULL;
 
-	return g_ptr_array_index (arch->rti_stack, arch->rti_stack->len - 1);
+	return g_ptr_array_index (arch->callback_stack, arch->callback_stack->len - 1);
 }
 
 static ServerCommandError
@@ -384,11 +411,11 @@ x86_arch_get_registers (ServerHandle *handle)
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
-	_server_ptrace_get_fp_registers (handle->inferior, &handle->arch->current_fpregs);
+	result = _server_ptrace_get_fp_registers (handle->inferior, &handle->arch->current_fpregs);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
-	_server_ptrace_get_dr (handle->inferior, DR_STATUS, &handle->arch->dr_status);
+	result = _server_ptrace_get_dr (handle->inferior, DR_STATUS, &handle->arch->dr_status);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
@@ -397,11 +424,14 @@ x86_arch_get_registers (ServerHandle *handle)
 
 ChildStoppedAction
 x86_arch_child_stopped (ServerHandle *handle, int stopsig,
-			guint64 *callback_arg, guint64 *retval, guint64 *retval2)
+			guint64 *callback_arg, guint64 *retval, guint64 *retval2,
+			guint32 *opt_data_size, gpointer *opt_data)
 {
 	ArchInfo *arch = handle->arch;
 	InferiorHandle *inferior = handle->inferior;
-	RuntimeInvokeData *rdata;
+	CodeBufferData *cbuffer = NULL;
+	CallbackData *cdata;
+	guint64 code;
 	int i;
 
 	x86_arch_get_registers (handle);
@@ -409,7 +439,8 @@ x86_arch_child_stopped (ServerHandle *handle, int stopsig,
 	if (stopsig == SIGSTOP)
 		return STOP_ACTION_INTERRUPTED;
 
-	if ((guint32) INFERIOR_REG_EIP (arch->current_regs) - 1 == inferior->notification_address) {
+	if (handle->mono_runtime &&
+	    ((guint32) INFERIOR_REG_EIP (arch->current_regs) - 1 == handle->mono_runtime->notification_address)) {
 		guint32 addr = (guint32) INFERIOR_REG_ESP (arch->current_regs) + 4;
 		guint64 data [3];
 
@@ -438,84 +469,108 @@ x86_arch_child_stopped (ServerHandle *handle, int stopsig,
 		return STOP_ACTION_BREAKPOINT_HIT;
 	}
 
-	rdata = get_runtime_invoke_data (arch);
-	if (rdata && (rdata->call_address == INFERIOR_REG_EIP (arch->current_regs))) {
+	cdata = get_callback_data (arch);
+	if (cdata && (cdata->call_address == INFERIOR_REG_EIP (arch->current_regs))) {
 		guint64 exc_object;
 
-		if (_server_ptrace_set_registers (inferior, rdata->saved_regs) != COMMAND_ERROR_NONE)
+		if (cdata->pushed_registers) {
+			guint32 pushed_regs [9];
+
+			if (_server_ptrace_read_memory (handle, cdata->pushed_registers, 36, &pushed_regs))
+				g_error (G_STRLOC ": Can't restore registers after returning from a call");
+
+			INFERIOR_REG_EAX (cdata->saved_regs) = pushed_regs [0];
+			INFERIOR_REG_EBX (cdata->saved_regs) = pushed_regs [1];
+			INFERIOR_REG_ECX (cdata->saved_regs) = pushed_regs [2];
+			INFERIOR_REG_EDX (cdata->saved_regs) = pushed_regs [3];
+			INFERIOR_REG_EBP (cdata->saved_regs) = pushed_regs [4];
+			INFERIOR_REG_ESP (cdata->saved_regs) = pushed_regs [5];
+			INFERIOR_REG_ESI (cdata->saved_regs) = pushed_regs [6];
+			INFERIOR_REG_EDI (cdata->saved_regs) = pushed_regs [7];
+			INFERIOR_REG_EIP (cdata->saved_regs) = pushed_regs [8];
+		}
+
+		if (_server_ptrace_set_registers (inferior, &cdata->saved_regs) != COMMAND_ERROR_NONE)
 			g_error (G_STRLOC ": Can't restore registers after returning from a call");
 
-		if (_server_ptrace_set_fp_registers (inferior, rdata->saved_fpregs) != COMMAND_ERROR_NONE)
+		if (_server_ptrace_set_fp_registers (inferior, &cdata->saved_fpregs) != COMMAND_ERROR_NONE)
 			g_error (G_STRLOC ": Can't restore FP registers after returning from a call");
 
-		*callback_arg = rdata->callback_argument;
+		*callback_arg = cdata->callback_argument;
 		*retval = (guint32) INFERIOR_REG_EAX (arch->current_regs);
 
-		if (rdata->exc_address &&
-		    (server_ptrace_peek_word (handle, rdata->exc_address, &exc_object) != COMMAND_ERROR_NONE))
+		if (cdata->data_pointer) {
+			*opt_data_size = cdata->data_size;
+			*opt_data = g_malloc0 (cdata->data_size);
+
+			if (_server_ptrace_read_memory (
+				    handle, cdata->data_pointer, cdata->data_size, *opt_data))
+				g_error (G_STRLOC ": Can't read data buffer after returning from a call");
+		} else {
+			*opt_data_size = 0;
+			*opt_data = NULL;
+		}
+
+
+		if (cdata->exc_address &&
+		    (server_ptrace_peek_word (handle, cdata->exc_address, &exc_object) != COMMAND_ERROR_NONE))
 			g_error (G_STRLOC ": Can't get exc object");
 
 		*retval2 = (guint32) exc_object;
 
-		inferior->last_signal = rdata->saved_signal;
-		g_free (rdata->saved_regs);
-		g_free (rdata->saved_fpregs);
-		g_ptr_array_remove (arch->rti_stack, rdata);
+		inferior->last_signal = cdata->saved_signal;
+		g_ptr_array_remove (arch->callback_stack, cdata);
 
 		x86_arch_get_registers (handle);
 
-		if (rdata->debug) {
+		if (cdata->debug) {
 			*retval = 0;
-			g_free (rdata);
+			g_free (cdata);
 			return STOP_ACTION_CALLBACK_COMPLETED;
 		}
 
-		g_free (rdata);
+		g_free (cdata);
 		return STOP_ACTION_CALLBACK;
 	}
 
-	if (!arch->call_address || arch->call_address != INFERIOR_REG_EIP (arch->current_regs)) {
-		guint64 code;
+	cbuffer = arch->code_buffer;
+	if (cbuffer) {
+		handle->mono_runtime->executable_code_bitfield [cbuffer->slot] = 0;
 
-#if defined(__linux__) || defined(__FreeBSD__)
-		if (stopsig != SIGTRAP)
+		if (cbuffer->code_address + cbuffer->insn_size != INFERIOR_REG_EIP (arch->current_regs)) {
+			g_warning (G_STRLOC ": %x - %x,%d - %x - %x", cbuffer->original_eip,
+				   cbuffer->code_address, cbuffer->insn_size,
+				   cbuffer->code_address + cbuffer->insn_size,
+				   INFERIOR_REG_EIP (arch->current_regs));
 			return STOP_ACTION_STOPPED;
-#endif
-
-		if (server_ptrace_peek_word (handle, GPOINTER_TO_SIZE (INFERIOR_REG_EIP (arch->current_regs) - 1), &code) != COMMAND_ERROR_NONE)
-			return STOP_ACTION_STOPPED;
-
-		if ((code & 0xff) == 0xcc) {
-			*retval = 0;
-			return STOP_ACTION_BREAKPOINT_HIT;
 		}
 
+		INFERIOR_REG_EIP (arch->current_regs) = cbuffer->original_eip;
+		if (cbuffer->update_ip)
+			INFERIOR_REG_EIP (arch->current_regs) += cbuffer->insn_size;
+		if (_server_ptrace_set_registers (inferior, &arch->current_regs) != COMMAND_ERROR_NONE) {
+			g_error (G_STRLOC ": Can't restore registers");
+		}
+
+		g_free (cbuffer);
+		arch->code_buffer = NULL;
 		return STOP_ACTION_STOPPED;
 	}
 
-	if (_server_ptrace_set_registers (inferior, arch->saved_regs) != COMMAND_ERROR_NONE)
-		g_error (G_STRLOC ": Can't restore registers after returning from a call");
+#if defined(__linux__) || defined(__FreeBSD__)
+	if (stopsig != SIGTRAP)
+		return STOP_ACTION_STOPPED;
+#endif
 
-	if (_server_ptrace_set_fp_registers (inferior, arch->saved_fpregs) != COMMAND_ERROR_NONE)
-		g_error (G_STRLOC ": Can't restore FP registers after returning from a call");
+	if (server_ptrace_peek_word (handle, GPOINTER_TO_SIZE (INFERIOR_REG_EIP (arch->current_regs) - 1), &code) != COMMAND_ERROR_NONE)
+		return STOP_ACTION_STOPPED;
 
-	*callback_arg = arch->callback_argument;
-	*retval = (guint32) INFERIOR_REG_EAX (arch->current_regs);
-	*retval2 = (guint32) INFERIOR_REG_EDX (arch->current_regs);
+	if ((code & 0xff) == 0xcc) {
+		*retval = 0;
+		return STOP_ACTION_BREAKPOINT_HIT;
+	}
 
-	inferior->last_signal = arch->saved_signal;
-	g_free (arch->saved_regs);
-	g_free (arch->saved_fpregs);
-
-	arch->saved_regs = NULL;
-	arch->saved_fpregs = NULL;
-	arch->saved_signal = 0;
-	arch->call_address = 0;
-	arch->callback_argument = 0;
-
-	x86_arch_get_registers (handle);
-
-	return STOP_ACTION_CALLBACK;
+	return STOP_ACTION_STOPPED;
 }
 
 static ServerCommandError
@@ -535,21 +590,21 @@ server_ptrace_get_registers (ServerHandle *handle, guint64 *values)
 {
 	ArchInfo *arch = handle->arch;
 
-	values [DEBUGGER_REG_EBX] = (guint32) INFERIOR_REG_EBX (arch->current_regs);
-	values [DEBUGGER_REG_ECX] = (guint32) INFERIOR_REG_ECX (arch->current_regs);
-	values [DEBUGGER_REG_EDX] = (guint32) INFERIOR_REG_EDX (arch->current_regs);
-	values [DEBUGGER_REG_ESI] = (guint32) INFERIOR_REG_ESI (arch->current_regs);
-	values [DEBUGGER_REG_EDI] = (guint32) INFERIOR_REG_EDI (arch->current_regs);
-	values [DEBUGGER_REG_EBP] = (guint32) INFERIOR_REG_EBP (arch->current_regs);
-	values [DEBUGGER_REG_EAX] = (guint32) INFERIOR_REG_EAX (arch->current_regs);
+	values [DEBUGGER_REG_RBX] = (guint32) INFERIOR_REG_EBX (arch->current_regs);
+	values [DEBUGGER_REG_RCX] = (guint32) INFERIOR_REG_ECX (arch->current_regs);
+	values [DEBUGGER_REG_RDX] = (guint32) INFERIOR_REG_EDX (arch->current_regs);
+	values [DEBUGGER_REG_RSI] = (guint32) INFERIOR_REG_ESI (arch->current_regs);
+	values [DEBUGGER_REG_RDI] = (guint32) INFERIOR_REG_EDI (arch->current_regs);
+	values [DEBUGGER_REG_RBP] = (guint32) INFERIOR_REG_EBP (arch->current_regs);
+	values [DEBUGGER_REG_RAX] = (guint32) INFERIOR_REG_EAX (arch->current_regs);
 	values [DEBUGGER_REG_DS] = (guint32) INFERIOR_REG_DS (arch->current_regs);
 	values [DEBUGGER_REG_ES] = (guint32) INFERIOR_REG_ES (arch->current_regs);
 	values [DEBUGGER_REG_FS] = (guint32) INFERIOR_REG_FS (arch->current_regs);
 	values [DEBUGGER_REG_GS] = (guint32) INFERIOR_REG_GS (arch->current_regs);
-	values [DEBUGGER_REG_EIP] = (guint32) INFERIOR_REG_EIP (arch->current_regs);
+	values [DEBUGGER_REG_RIP] = (guint32) INFERIOR_REG_EIP (arch->current_regs);
 	values [DEBUGGER_REG_CS] = (guint32) INFERIOR_REG_CS (arch->current_regs);
 	values [DEBUGGER_REG_EFLAGS] = (guint32) INFERIOR_REG_EFLAGS (arch->current_regs);
-	values [DEBUGGER_REG_ESP] = (guint32) INFERIOR_REG_ESP (arch->current_regs);
+	values [DEBUGGER_REG_RSP] = (guint32) INFERIOR_REG_ESP (arch->current_regs);
 	values [DEBUGGER_REG_SS] = (guint32) INFERIOR_REG_SS (arch->current_regs);
 
 	return COMMAND_ERROR_NONE;
@@ -560,21 +615,21 @@ server_ptrace_set_registers (ServerHandle *handle, guint64 *values)
 {
 	ArchInfo *arch = handle->arch;
 
-	INFERIOR_REG_EBX (arch->current_regs) = values [DEBUGGER_REG_EBX];
-	INFERIOR_REG_ECX (arch->current_regs) = values [DEBUGGER_REG_ECX];
-	INFERIOR_REG_EDX (arch->current_regs) = values [DEBUGGER_REG_EDX];
-	INFERIOR_REG_ESI (arch->current_regs) = values [DEBUGGER_REG_ESI];
-	INFERIOR_REG_EDI (arch->current_regs) = values [DEBUGGER_REG_EDI];
-	INFERIOR_REG_EBP (arch->current_regs) = values [DEBUGGER_REG_EBP];
-	INFERIOR_REG_EAX (arch->current_regs) = values [DEBUGGER_REG_EAX];
+	INFERIOR_REG_EBX (arch->current_regs) = values [DEBUGGER_REG_RBX];
+	INFERIOR_REG_ECX (arch->current_regs) = values [DEBUGGER_REG_RCX];
+	INFERIOR_REG_EDX (arch->current_regs) = values [DEBUGGER_REG_RDX];
+	INFERIOR_REG_ESI (arch->current_regs) = values [DEBUGGER_REG_RSI];
+	INFERIOR_REG_EDI (arch->current_regs) = values [DEBUGGER_REG_RDI];
+	INFERIOR_REG_EBP (arch->current_regs) = values [DEBUGGER_REG_RBP];
+	INFERIOR_REG_EAX (arch->current_regs) = values [DEBUGGER_REG_RAX];
 	INFERIOR_REG_DS (arch->current_regs) = values [DEBUGGER_REG_DS];
 	INFERIOR_REG_ES (arch->current_regs) = values [DEBUGGER_REG_ES];
 	INFERIOR_REG_FS (arch->current_regs) = values [DEBUGGER_REG_FS];
 	INFERIOR_REG_GS (arch->current_regs) = values [DEBUGGER_REG_GS];
-	INFERIOR_REG_EIP (arch->current_regs) = values [DEBUGGER_REG_EIP];
+	INFERIOR_REG_EIP (arch->current_regs) = values [DEBUGGER_REG_RIP];
 	INFERIOR_REG_CS (arch->current_regs) = values [DEBUGGER_REG_CS];
 	INFERIOR_REG_EFLAGS (arch->current_regs) = values [DEBUGGER_REG_EFLAGS];
-	INFERIOR_REG_ESP (arch->current_regs) = values [DEBUGGER_REG_ESP];
+	INFERIOR_REG_ESP (arch->current_regs) = values [DEBUGGER_REG_RSP];
 	INFERIOR_REG_SS (arch->current_regs) = values [DEBUGGER_REG_SS];
 
 	return _server_ptrace_set_registers (handle->inferior, &arch->current_regs);
@@ -615,8 +670,82 @@ server_ptrace_pop_registers (ServerHandle *handle)
 	return COMMAND_ERROR_NONE;
 }
 
+static int
+find_breakpoint_table_slot (MonoRuntimeInfo *runtime)
+{
+	int i;
+
+	for (i = 1; i < runtime->breakpoint_table_size; i++) {
+		if (runtime->breakpoint_table_bitfield [i])
+			continue;
+
+		runtime->breakpoint_table_bitfield [i] = 1;
+		return i;
+	}
+
+	return -1;
+}
+
 static ServerCommandError
-do_enable (ServerHandle *handle, BreakpointInfo *breakpoint)
+runtime_info_enable_breakpoint (ServerHandle *handle, BreakpointInfo *breakpoint)
+{
+	MonoRuntimeInfo *runtime;
+	ServerCommandError result;
+	guint64 table_address, index_address;
+	int slot;
+
+	runtime = handle->mono_runtime;
+	g_assert (runtime);
+
+	slot = find_breakpoint_table_slot (runtime);
+	if (slot < 0)
+		return COMMAND_ERROR_INTERNAL_ERROR;
+
+	breakpoint->runtime_table_slot = slot;
+
+	table_address = runtime->breakpoint_info_area + 8 * slot;
+	index_address = runtime->breakpoint_table + 4 * slot;
+
+	result = server_ptrace_poke_word (handle, table_address, breakpoint->address);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	result = server_ptrace_poke_word (handle, table_address + 4, (gsize) breakpoint->saved_insn);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	result = server_ptrace_poke_word (handle, index_address, (gsize) slot);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	return COMMAND_ERROR_NONE;
+}
+
+static ServerCommandError
+runtime_info_disable_breakpoint (ServerHandle *handle, BreakpointInfo *breakpoint)
+{
+	MonoRuntimeInfo *runtime;
+	ServerCommandError result;
+	guint64 index_address;
+	int slot;
+
+	runtime = handle->mono_runtime;
+	g_assert (runtime);
+
+	slot = breakpoint->runtime_table_slot;
+	index_address = runtime->breakpoint_table + runtime->address_size * slot;
+
+	result = server_ptrace_poke_word (handle, index_address, 0);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	runtime->breakpoint_table_bitfield [slot] = 0;
+
+	return COMMAND_ERROR_NONE;
+}
+
+static ServerCommandError
+x86_arch_enable_breakpoint (ServerHandle *handle, BreakpointInfo *breakpoint)
 {
 	ServerCommandError result;
 	ArchInfo *arch = handle->arch;
@@ -651,6 +780,12 @@ do_enable (ServerHandle *handle, BreakpointInfo *breakpoint)
 		if (result != COMMAND_ERROR_NONE)
 			return result;
 
+		if (handle->mono_runtime) {
+			result = runtime_info_enable_breakpoint (handle, breakpoint);
+			if (result != COMMAND_ERROR_NONE)
+				return result;
+		}
+
 		result = server_ptrace_write_memory (handle, address, 1, &bopcode);
 		if (result != COMMAND_ERROR_NONE)
 			return result;
@@ -660,7 +795,7 @@ do_enable (ServerHandle *handle, BreakpointInfo *breakpoint)
 }
 
 static ServerCommandError
-do_disable (ServerHandle *handle, BreakpointInfo *breakpoint)
+x86_arch_disable_breakpoint (ServerHandle *handle, BreakpointInfo *breakpoint)
 {
 	ServerCommandError result;
 	ArchInfo *arch = handle->arch;
@@ -692,6 +827,12 @@ do_disable (ServerHandle *handle, BreakpointInfo *breakpoint)
 		result = server_ptrace_write_memory (handle, address, 1, &breakpoint->saved_insn);
 		if (result != COMMAND_ERROR_NONE)
 			return result;
+
+		if (handle->mono_runtime) {
+			result = runtime_info_disable_breakpoint (handle, breakpoint);
+			if (result != COMMAND_ERROR_NONE)
+				return result;
+		}
 	}
 
 	return COMMAND_ERROR_NONE;
@@ -727,7 +868,7 @@ server_ptrace_insert_breakpoint (ServerHandle *handle, guint64 address, guint32
 	breakpoint->id = mono_debugger_breakpoint_manager_get_next_id ();
 	breakpoint->dr_index = -1;
 
-	result = do_enable (handle, breakpoint);
+	result = x86_arch_enable_breakpoint (handle, breakpoint);
 	if (result != COMMAND_ERROR_NONE) {
 		mono_debugger_breakpoint_manager_unlock ();
 		g_free (breakpoint);
@@ -761,7 +902,7 @@ server_ptrace_remove_breakpoint (ServerHandle *handle, guint32 bhandle)
 		goto out;
 	}
 
-	result = do_disable (handle, breakpoint);
+	result = x86_arch_disable_breakpoint (handle, breakpoint);
 	if (result != COMMAND_ERROR_NONE)
 		goto out;
 
@@ -788,26 +929,6 @@ x86_arch_remove_hardware_breakpoints (ServerHandle *handle)
 	}
 }
 
-static ServerCommandError
-server_ptrace_init_after_fork (ServerHandle *handle)
-{
-	GPtrArray *breakpoints;
-	int i;
-
-	mono_debugger_breakpoint_manager_lock ();
-
-	breakpoints = mono_debugger_breakpoint_manager_get_breakpoints (handle->bpm);
-	for (i = 0; i < breakpoints->len; i++) {
-		BreakpointInfo *info = g_ptr_array_index (breakpoints, i);
-
-		if (info->dr_index >= 0)
-			do_disable (handle, info);
-	}
-
-	mono_debugger_breakpoint_manager_unlock ();
-	return COMMAND_ERROR_NONE;
-}
-
 static ServerCommandError
 find_free_hw_register (ServerHandle *handle, guint32 *idx)
 {
@@ -851,7 +972,7 @@ server_ptrace_insert_hw_breakpoint (ServerHandle *handle, guint32 type, guint32
 	breakpoint->is_hardware_bpt = TRUE;
 	breakpoint->dr_index = *idx;
 
-	result = do_enable (handle, breakpoint);
+	result = x86_arch_enable_breakpoint (handle, breakpoint);
 	if (result != COMMAND_ERROR_NONE) {
 		mono_debugger_breakpoint_manager_unlock ();
 		g_free (breakpoint);
@@ -880,7 +1001,7 @@ server_ptrace_enable_breakpoint (ServerHandle *handle, guint32 bhandle)
 		return COMMAND_ERROR_NO_SUCH_BREAKPOINT;
 	}
 
-	result = do_enable (handle, breakpoint);
+	result = x86_arch_enable_breakpoint (handle, breakpoint);
 	breakpoint->enabled = TRUE;
 	mono_debugger_breakpoint_manager_unlock ();
 	return result;
@@ -899,7 +1020,7 @@ server_ptrace_disable_breakpoint (ServerHandle *handle, guint32 bhandle)
 		return COMMAND_ERROR_NO_SUCH_BREAKPOINT;
 	}
 
-	result = do_disable (handle, breakpoint);
+	result = x86_arch_disable_breakpoint (handle, breakpoint);
 	breakpoint->enabled = FALSE;
 	mono_debugger_breakpoint_manager_unlock ();
 	return result;
@@ -926,29 +1047,159 @@ server_ptrace_get_breakpoints (ServerHandle *handle, guint32 *count, guint32 **r
 	return COMMAND_ERROR_NONE;	
 }
 
+static int
+find_code_buffer_slot (MonoRuntimeInfo *runtime)
+{
+	int i;
+
+	for (i = 0; i < runtime->executable_code_total_chunks; i++) {
+		if (runtime->executable_code_bitfield [i])
+			continue;
+
+		runtime->executable_code_bitfield [i] = 1;
+		return i;
+	}
+
+	return -1;
+}
 
 static ServerCommandError
-server_ptrace_abort_invoke (ServerHandle *handle)
+server_ptrace_execute_instruction (ServerHandle *handle, const guint8 *instruction,
+				   guint32 size, gboolean update_ip)
 {
-	RuntimeInvokeData *rdata;
+	MonoRuntimeInfo *runtime;
+	ServerCommandError result;
+	CodeBufferData *data;
+	guint32 code_address;
+	int slot;
 
-	rdata = get_runtime_invoke_data (handle->arch);
-	if (!rdata)
-		return COMMAND_ERROR_UNKNOWN_ERROR;
+	runtime = handle->mono_runtime;
+	g_assert (runtime);
 
-	if (_server_ptrace_set_registers (handle->inferior, rdata->saved_regs) != COMMAND_ERROR_NONE)
+	if (!runtime->executable_code_buffer)
+		return COMMAND_ERROR_INTERNAL_ERROR;
+
+	slot = find_code_buffer_slot (runtime);
+	if (slot < 0)
+		return COMMAND_ERROR_INTERNAL_ERROR;
+
+	if (size > runtime->executable_code_chunk_size)
+		return COMMAND_ERROR_INTERNAL_ERROR;
+	if (handle->arch->code_buffer)
+		return COMMAND_ERROR_INTERNAL_ERROR;
+
+	code_address = runtime->executable_code_buffer + slot * runtime->executable_code_chunk_size;
+
+	data = g_new0 (CodeBufferData, 1);
+	data->slot = slot;
+	data->insn_size = size;
+	data->update_ip = update_ip;
+	data->original_eip = INFERIOR_REG_EIP (handle->arch->current_regs);
+	data->code_address = code_address;
+
+	handle->arch->code_buffer = data;
+
+	result = server_ptrace_write_memory (handle, code_address, size, instruction);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	INFERIOR_REG_EIP (handle->arch->current_regs) = code_address;
+
+	result = _server_ptrace_set_registers (handle->inferior, &handle->arch->current_regs);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	return server_ptrace_step (handle);
+}
+
+static ServerCommandError
+server_ptrace_mark_rti_frame (ServerHandle *handle)
+{
+	CallbackData *cdata;
+
+	cdata = get_callback_data (handle->arch);
+	if (!cdata)
+		return COMMAND_ERROR_NO_CALLBACK_FRAME;
+
+	cdata->rti_frame = INFERIOR_REG_ESP (handle->arch->current_regs) + 4;
+	return COMMAND_ERROR_NONE;
+}
+
+static ServerCommandError
+server_ptrace_abort_invoke (ServerHandle *handle, guint64 stack_pointer)
+{
+	CallbackData *cdata;
+
+	cdata = get_callback_data (handle->arch);
+	if (!cdata)
+		return COMMAND_ERROR_NO_CALLBACK_FRAME;
+
+	if (cdata->rti_frame && (stack_pointer < cdata->rti_frame))
+		return COMMAND_ERROR_NO_CALLBACK_FRAME;
+	if (stack_pointer < cdata->stack_pointer)
+		return COMMAND_ERROR_NO_CALLBACK_FRAME;
+
+	if (_server_ptrace_set_registers (handle->inferior, &cdata->saved_regs) != COMMAND_ERROR_NONE)
 		g_error (G_STRLOC ": Can't restore registers after returning from a call");
 
-	if (_server_ptrace_set_fp_registers (handle->inferior, rdata->saved_fpregs) != COMMAND_ERROR_NONE)
+	if (_server_ptrace_set_fp_registers (handle->inferior, &cdata->saved_fpregs) != COMMAND_ERROR_NONE)
 		g_error (G_STRLOC ": Can't restore FP registers after returning from a call");
 
-	handle->inferior->last_signal = rdata->saved_signal;
-	g_free (rdata->saved_regs);
-	g_free (rdata->saved_fpregs);
-	g_ptr_array_remove (handle->arch->rti_stack, rdata);
+	handle->inferior->last_signal = cdata->saved_signal;
+	g_ptr_array_remove (handle->arch->callback_stack, cdata);
 
 	x86_arch_get_registers (handle);
-	g_free (rdata);
+	g_free (cdata);
 
 	return COMMAND_ERROR_NONE;
 }
+
+static ServerCommandError
+server_ptrace_get_callback_frame (ServerHandle *handle, guint64 stack_pointer,
+				  gboolean exact_match, guint64 *registers)
+{
+	int i;
+
+	for (i = 0; i < handle->arch->callback_stack->len; i++) {
+		CallbackData *cdata = g_ptr_array_index (handle->arch->callback_stack, i);
+
+		if (cdata->rti_frame) {
+			if (exact_match) {
+				if (cdata->rti_frame != stack_pointer)
+					continue;
+			} else {
+				if (cdata->rti_frame < stack_pointer)
+					continue;
+			}
+		} else {
+			if (exact_match) {
+				if (cdata->stack_pointer != stack_pointer)
+					continue;
+			} else {
+				if (cdata->stack_pointer < stack_pointer)
+					continue;
+			}
+		}
+
+		registers [DEBUGGER_REG_RBX] = (guint32) INFERIOR_REG_EBX (cdata->saved_regs);
+		registers [DEBUGGER_REG_RCX] = (guint32) INFERIOR_REG_ECX (cdata->saved_regs);
+		registers [DEBUGGER_REG_RDX] = (guint32) INFERIOR_REG_EDX (cdata->saved_regs);
+		registers [DEBUGGER_REG_RSI] = (guint32) INFERIOR_REG_ESI (cdata->saved_regs);
+		registers [DEBUGGER_REG_RDI] = (guint32) INFERIOR_REG_EDI (cdata->saved_regs);
+		registers [DEBUGGER_REG_RBP] = (guint32) INFERIOR_REG_EBP (cdata->saved_regs);
+		registers [DEBUGGER_REG_RAX] = (guint32) INFERIOR_REG_EAX (cdata->saved_regs);
+		registers [DEBUGGER_REG_DS] = (guint32) INFERIOR_REG_DS (cdata->saved_regs);
+		registers [DEBUGGER_REG_ES] = (guint32) INFERIOR_REG_ES (cdata->saved_regs);
+		registers [DEBUGGER_REG_FS] = (guint32) INFERIOR_REG_FS (cdata->saved_regs);
+		registers [DEBUGGER_REG_GS] = (guint32) INFERIOR_REG_GS (cdata->saved_regs);
+		registers [DEBUGGER_REG_RIP] = (guint32) INFERIOR_REG_EIP (cdata->saved_regs);
+		registers [DEBUGGER_REG_CS] = (guint32) INFERIOR_REG_CS (cdata->saved_regs);
+		registers [DEBUGGER_REG_EFLAGS] = (guint32) INFERIOR_REG_EFLAGS (cdata->saved_regs);
+		registers [DEBUGGER_REG_RSP] = (guint32) INFERIOR_REG_ESP (cdata->saved_regs);
+		registers [DEBUGGER_REG_SS] = (guint32) INFERIOR_REG_SS (cdata->saved_regs);
+
+		return COMMAND_ERROR_NONE;
+	}
+
+	return COMMAND_ERROR_NO_CALLBACK_FRAME;
+}
