@@ -30,27 +30,6 @@ G_BEGIN_DECLS
 #define INFERIOR_REG_SS(r)	r.ss
 #define INFERIOR_REG_GS(r)	r.gs
 
-typedef enum {
-	DEBUGGER_REG_EBX = 0,
-	DEBUGGER_REG_ECX,
-	DEBUGGER_REG_EDX,
-	DEBUGGER_REG_ESI,
-	DEBUGGER_REG_EDI,
-	DEBUGGER_REG_EBP,
-	DEBUGGER_REG_EAX,
-	DEBUGGER_REG_DS,
-	DEBUGGER_REG_ES,
-	DEBUGGER_REG_FS,
-	DEBUGGER_REG_GS,
-	DEBUGGER_REG_EIP,
-	DEBUGGER_REG_CS,
-	DEBUGGER_REG_EFLAGS,
-	DEBUGGER_REG_ESP,
-	DEBUGGER_REG_SS,
-
-	DEBUGGER_REG_LAST
-} DebuggerRegisters;
-
 G_END_DECLS
 
 #else
