@@ -3,6 +3,9 @@
 
 #include "x86-arch.h"
 
+static ServerCommandError
+_server_ptrace_check_errno (InferiorHandle *);
+
 static ServerCommandError
 _server_ptrace_get_registers (InferiorHandle *inferior, INFERIOR_REGS_TYPE *regs);
 
@@ -15,6 +18,10 @@ _server_ptrace_get_fp_registers (InferiorHandle *inferior, INFERIOR_FPREGS_TYPE
 static ServerCommandError
 _server_ptrace_set_fp_registers (InferiorHandle *inferior, INFERIOR_FPREGS_TYPE *regs);
 
+static ServerCommandError
+_server_ptrace_read_memory (ServerHandle *handle, guint64 start,
+			    guint32 size, gpointer buffer);
+
 static ServerCommandError
 server_ptrace_read_memory (ServerHandle *handle, guint64 start, guint32 size, gpointer buffer);
 
