@@ -15,16 +15,21 @@
 
 #define AMD64_RED_ZONE_SIZE 128
 
+typedef struct
+{
+	int slot;
+	int insn_size;
+	gboolean update_ip;
+	guint64 code_address;
+	guint64 original_rip;
+} CodeBufferData;
+
 struct ArchInfo
 {
-	guint64 call_address;
-	guint64 callback_argument;
 	INFERIOR_REGS_TYPE current_regs;
 	INFERIOR_FPREGS_TYPE current_fpregs;
-	INFERIOR_REGS_TYPE *saved_regs;
-	INFERIOR_FPREGS_TYPE *saved_fpregs;
-	int saved_signal;
-	GPtrArray *rti_stack;
+	GPtrArray *callback_stack;
+	CodeBufferData *code_buffer;
 	guint64 dr_control, dr_status;
 	guint64 pushed_regs_rsp;
 	int dr_regs [DR_NADDR];
@@ -32,21 +37,26 @@ struct ArchInfo
 
 typedef struct
 {
-	INFERIOR_REGS_TYPE *saved_regs;
-	INFERIOR_FPREGS_TYPE *saved_fpregs;
-	guint64 call_address;
-	guint64 exc_address;
+	INFERIOR_REGS_TYPE saved_regs;
+	INFERIOR_FPREGS_TYPE saved_fpregs;
 	guint64 callback_argument;
-	gboolean debug;
+	guint64 call_address;
+	guint64 stack_pointer;
+	guint64 rti_frame;
+	guint64 exc_address;
 	int saved_signal;
-} RuntimeInvokeData;
+	guint64 pushed_registers;
+	guint64 data_pointer;
+	guint32 data_size;
+	gboolean debug;
+} CallbackData;
 
 ArchInfo *
 x86_arch_initialize (void)
 {
 	ArchInfo *arch = g_new0 (ArchInfo, 1);
 
-	arch->rti_stack = g_ptr_array_new ();
+	arch->callback_stack = g_ptr_array_new ();
 
 	return arch;
 }
@@ -54,9 +64,7 @@ x86_arch_initialize (void)
 void
 x86_arch_finalize (ArchInfo *arch)
 {
-	g_ptr_array_free (arch->rti_stack, TRUE);
-	g_free (arch->saved_regs);
-     	g_free (arch->saved_fpregs);
+	g_ptr_array_free (arch->callback_stack, TRUE);
 	g_free (arch);
 }
 
@@ -132,13 +140,13 @@ check_breakpoint (ServerHandle *handle, guint64 address, guint64 *retval)
 	return TRUE;
 }
 
-static RuntimeInvokeData *
-get_runtime_invoke_data (ArchInfo *arch)
+static CallbackData *
+get_callback_data (ArchInfo *arch)
 {
-	if (!arch->rti_stack->len)
+	if (!arch->callback_stack->len)
 		return NULL;
 
-	return g_ptr_array_index (arch->rti_stack, arch->rti_stack->len - 1);
+	return g_ptr_array_index (arch->callback_stack, arch->callback_stack->len - 1);
 }
 
 static ServerCommandError
@@ -150,11 +158,11 @@ x86_arch_get_registers (ServerHandle *handle)
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
-	_server_ptrace_get_fp_registers (handle->inferior, &handle->arch->current_fpregs);
+	result = _server_ptrace_get_fp_registers (handle->inferior, &handle->arch->current_fpregs);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
-	_server_ptrace_get_dr (handle->inferior, DR_STATUS, &handle->arch->dr_status);
+	result = _server_ptrace_get_dr (handle->inferior, DR_STATUS, &handle->arch->dr_status);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
@@ -163,11 +171,14 @@ x86_arch_get_registers (ServerHandle *handle)
 
 ChildStoppedAction
 x86_arch_child_stopped (ServerHandle *handle, int stopsig,
-			guint64 *callback_arg, guint64 *retval, guint64 *retval2)
+			guint64 *callback_arg, guint64 *retval, guint64 *retval2,
+			guint32 *opt_data_size, gpointer *opt_data)
 {
 	ArchInfo *arch = handle->arch;
 	InferiorHandle *inferior = handle->inferior;
-	RuntimeInvokeData *rdata;
+	CodeBufferData *cbuffer = NULL;
+	CallbackData *cdata;
+	guint64 code;
 	int i;
 
 	x86_arch_get_registers (handle);
@@ -175,7 +186,8 @@ x86_arch_child_stopped (ServerHandle *handle, int stopsig,
 	if (stopsig == SIGSTOP)
 		return STOP_ACTION_INTERRUPTED;
 
-	if (INFERIOR_REG_RIP (arch->current_regs) - 1 == inferior->notification_address) {
+	if (handle->mono_runtime &&
+	    (INFERIOR_REG_RIP (arch->current_regs) - 1 == handle->mono_runtime->notification_address)) {
 		*callback_arg = INFERIOR_REG_RDI (arch->current_regs);
 		*retval = INFERIOR_REG_RSI (arch->current_regs);
 		*retval2 = INFERIOR_REG_RDX (arch->current_regs);
@@ -198,80 +210,107 @@ x86_arch_child_stopped (ServerHandle *handle, int stopsig,
 		return STOP_ACTION_BREAKPOINT_HIT;
 	}
 
-	rdata = get_runtime_invoke_data (arch);
-	if (rdata && (rdata->call_address == INFERIOR_REG_RIP (arch->current_regs))) {
-		if (_server_ptrace_set_registers (inferior, rdata->saved_regs) != COMMAND_ERROR_NONE)
+	cdata = get_callback_data (arch);
+	if (cdata && (cdata->call_address == INFERIOR_REG_RIP (arch->current_regs))) {
+		if (cdata->pushed_registers) {
+			guint64 pushed_regs [13];
+
+			if (_server_ptrace_read_memory (handle, cdata->pushed_registers, 104, &pushed_regs))
+				g_error (G_STRLOC ": Can't restore registers after returning from a call");
+
+			INFERIOR_REG_RAX (cdata->saved_regs) = pushed_regs [0];
+			INFERIOR_REG_RBX (cdata->saved_regs) = pushed_regs [1];
+			INFERIOR_REG_RCX (cdata->saved_regs) = pushed_regs [2];
+			INFERIOR_REG_RDX (cdata->saved_regs) = pushed_regs [3];
+			INFERIOR_REG_RBP (cdata->saved_regs) = pushed_regs [4];
+			INFERIOR_REG_RSP (cdata->saved_regs) = pushed_regs [5];
+			INFERIOR_REG_RSI (cdata->saved_regs) = pushed_regs [6];
+			INFERIOR_REG_RDI (cdata->saved_regs) = pushed_regs [7];
+			INFERIOR_REG_RIP (cdata->saved_regs) = pushed_regs [8];
+			INFERIOR_REG_R12 (cdata->saved_regs) = pushed_regs [9];
+			INFERIOR_REG_R13 (cdata->saved_regs) = pushed_regs [10];
+			INFERIOR_REG_R14 (cdata->saved_regs) = pushed_regs [11];
+			INFERIOR_REG_R15 (cdata->saved_regs) = pushed_regs [12];
+		}
+
+		if (_server_ptrace_set_registers (inferior, &cdata->saved_regs) != COMMAND_ERROR_NONE)
 			g_error (G_STRLOC ": Can't restore registers after returning from a call");
 
-		if (_server_ptrace_set_fp_registers (inferior, rdata->saved_fpregs) != COMMAND_ERROR_NONE)
+		if (_server_ptrace_set_fp_registers (inferior, &cdata->saved_fpregs) != COMMAND_ERROR_NONE)
 			g_error (G_STRLOC ": Can't restore FP registers after returning from a call");
 
-		*callback_arg = rdata->callback_argument;
+		*callback_arg = cdata->callback_argument;
 		*retval = INFERIOR_REG_RAX (arch->current_regs);
 
-		if (rdata->exc_address &&
-		    (server_ptrace_peek_word (handle, rdata->exc_address, retval2) != COMMAND_ERROR_NONE))
+		if (cdata->data_pointer) {
+			*opt_data_size = cdata->data_size;
+			*opt_data = g_malloc0 (cdata->data_size);
+
+			if (_server_ptrace_read_memory (
+				    handle, cdata->data_pointer, cdata->data_size, *opt_data))
+				g_error (G_STRLOC ": Can't read data buffer after returning from a call");
+		} else {
+			*opt_data_size = 0;
+			*opt_data = NULL;
+		}
+
+		if (cdata->exc_address &&
+		    (server_ptrace_peek_word (handle, cdata->exc_address, retval2) != COMMAND_ERROR_NONE))
 			g_error (G_STRLOC ": Can't get exc object");
 
-		inferior->last_signal = rdata->saved_signal;
-		g_free (rdata->saved_regs);
-		g_free (rdata->saved_fpregs);
-		g_ptr_array_remove (arch->rti_stack, rdata);
+		inferior->last_signal = cdata->saved_signal;
+		g_ptr_array_remove (arch->callback_stack, cdata);
 
 		x86_arch_get_registers (handle);
 
-		if (rdata->debug) {
+		if (cdata->debug) {
 			*retval = 0;
-			g_free (rdata);
+			g_free (cdata);
 			return STOP_ACTION_CALLBACK_COMPLETED;
 		}
 
-		g_free (rdata);
+		g_free (cdata);
 		return STOP_ACTION_CALLBACK;
 	}
 
-	if (!arch->call_address || arch->call_address != INFERIOR_REG_RIP (arch->current_regs)) {
-		guint64 code;
+	cbuffer = arch->code_buffer;
+	if (cbuffer) {
+		handle->mono_runtime->executable_code_bitfield [cbuffer->slot] = 0;
 
-#if defined(__linux__) || defined(__FreeBSD__)
-		if (stopsig != SIGTRAP)
+		if (cbuffer->code_address + cbuffer->insn_size != INFERIOR_REG_RIP (arch->current_regs)) {
+			g_warning (G_STRLOC ": %Lx,%d - %Lx - %Lx",
+				   cbuffer->code_address, cbuffer->insn_size,
+				   cbuffer->code_address + cbuffer->insn_size,
+				   INFERIOR_REG_RIP (arch->current_regs));
 			return STOP_ACTION_STOPPED;
-#endif
-
-		if (server_ptrace_peek_word (handle, GPOINTER_TO_SIZE(INFERIOR_REG_RIP (arch->current_regs) - 1), &code) != COMMAND_ERROR_NONE)
-			return STOP_ACTION_STOPPED;
-
-		if ((code & 0xff) == 0xcc) {
-			*retval = 0;
-			return STOP_ACTION_BREAKPOINT_HIT;
 		}
 
+		INFERIOR_REG_RIP (arch->current_regs) = cbuffer->original_rip;
+		if (cbuffer->update_ip)
+			INFERIOR_REG_RIP (arch->current_regs) += cbuffer->insn_size;
+		if (_server_ptrace_set_registers (inferior, &arch->current_regs) != COMMAND_ERROR_NONE) {
+			g_error (G_STRLOC ": Can't restore registers");
+		}
+
+		g_free (cbuffer);
+		arch->code_buffer = NULL;
 		return STOP_ACTION_STOPPED;
 	}
 
-	if (_server_ptrace_set_registers (inferior, arch->saved_regs) != COMMAND_ERROR_NONE)
-		g_error (G_STRLOC ": Can't restore registers after returning from a call");
+#if defined(__linux__) || defined(__FreeBSD__)
+	if (stopsig != SIGTRAP)
+		return STOP_ACTION_STOPPED;
+#endif
 
-	if (_server_ptrace_set_fp_registers (inferior, arch->saved_fpregs) != COMMAND_ERROR_NONE)
-		g_error (G_STRLOC ": Can't restore FP registers after returning from a call");
+	if (server_ptrace_peek_word (handle, GPOINTER_TO_SIZE(INFERIOR_REG_RIP (arch->current_regs) - 1), &code) != COMMAND_ERROR_NONE)
+		return STOP_ACTION_STOPPED;
 
-	*callback_arg = arch->callback_argument;
-	*retval = INFERIOR_REG_RAX (arch->current_regs);
-	*retval2 = 0;
+	if ((code & 0xff) == 0xcc) {
+		*retval = 0;
+		return STOP_ACTION_BREAKPOINT_HIT;
+	}
 
-	inferior->last_signal = arch->saved_signal;
-	g_free (arch->saved_regs);
-	g_free (arch->saved_fpregs);
-
-	arch->saved_regs = NULL;
-	arch->saved_fpregs = NULL;
-	arch->saved_signal = 0;
-	arch->call_address = 0;
-	arch->callback_argument = 0;
-
-	x86_arch_get_registers (handle);
-
-	return STOP_ACTION_CALLBACK;
+	return STOP_ACTION_STOPPED;
 }
 
 static ServerCommandError
@@ -405,8 +444,84 @@ server_ptrace_pop_registers (ServerHandle *handle)
 	return COMMAND_ERROR_NONE;
 }
 
+static int
+find_breakpoint_table_slot (MonoRuntimeInfo *runtime)
+{
+	int i;
+
+	for (i = 1; i < runtime->breakpoint_table_size; i++) {
+		if (runtime->breakpoint_table_bitfield [i])
+			continue;
+
+		runtime->breakpoint_table_bitfield [i] = 1;
+		return i;
+	}
+
+	return -1;
+}
+
 static ServerCommandError
-do_enable (ServerHandle *handle, BreakpointInfo *breakpoint)
+runtime_info_enable_breakpoint (ServerHandle *handle, BreakpointInfo *breakpoint)
+{
+	MonoRuntimeInfo *runtime;
+	ServerCommandError result;
+	guint64 table_address, index_address;
+	int slot;
+
+	runtime = handle->mono_runtime;
+	g_assert (runtime);
+
+	slot = find_breakpoint_table_slot (runtime);
+	if (slot < 0)
+		return COMMAND_ERROR_INTERNAL_ERROR;
+
+	breakpoint->runtime_table_slot = slot;
+
+	table_address = runtime->breakpoint_info_area + 16 * slot;
+	index_address = runtime->breakpoint_table + 8 * slot;
+
+	result = server_ptrace_poke_word (handle, table_address, breakpoint->address);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	result = server_ptrace_poke_word (handle, table_address + 8, (gsize) breakpoint->saved_insn);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	result = server_ptrace_poke_word (handle, index_address, (gsize) slot);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	return COMMAND_ERROR_NONE;
+}
+
+static ServerCommandError
+runtime_info_disable_breakpoint (ServerHandle *handle, BreakpointInfo *breakpoint)
+{
+	MonoRuntimeInfo *runtime;
+	ServerCommandError result;
+	guint64 index_address;
+	int slot;
+
+	runtime = handle->mono_runtime;
+	g_assert (runtime);
+
+	return COMMAND_ERROR_NONE;
+
+	slot = breakpoint->runtime_table_slot;
+	index_address = runtime->breakpoint_table + runtime->address_size * slot;
+
+	result = server_ptrace_poke_word (handle, index_address, 0);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	runtime->breakpoint_table_bitfield [slot] = 0;
+
+	return COMMAND_ERROR_NONE;
+}
+
+static ServerCommandError
+x86_arch_enable_breakpoint (ServerHandle *handle, BreakpointInfo *breakpoint)
 {
 	ServerCommandError result;
 	ArchInfo *arch = handle->arch;
@@ -446,6 +561,12 @@ do_enable (ServerHandle *handle, BreakpointInfo *breakpoint)
 		if (result != COMMAND_ERROR_NONE)
 			return result;
 
+		if (handle->mono_runtime) {
+			result = runtime_info_enable_breakpoint (handle, breakpoint);
+			if (result != COMMAND_ERROR_NONE)
+				return result;
+		}
+
 		result = server_ptrace_write_memory (handle, address, 1, &bopcode);
 		if (result != COMMAND_ERROR_NONE)
 			return result;
@@ -455,7 +576,7 @@ do_enable (ServerHandle *handle, BreakpointInfo *breakpoint)
 }
 
 static ServerCommandError
-do_disable (ServerHandle *handle, BreakpointInfo *breakpoint)
+x86_arch_disable_breakpoint (ServerHandle *handle, BreakpointInfo *breakpoint)
 {
 	ServerCommandError result;
 	ArchInfo *arch = handle->arch;
@@ -487,6 +608,12 @@ do_disable (ServerHandle *handle, BreakpointInfo *breakpoint)
 		result = server_ptrace_write_memory (handle, address, 1, &breakpoint->saved_insn);
 		if (result != COMMAND_ERROR_NONE)
 			return result;
+
+		if (handle->mono_runtime) {
+			result = runtime_info_disable_breakpoint (handle, breakpoint);
+			if (result != COMMAND_ERROR_NONE)
+				return result;
+		}
 	}
 
 	return COMMAND_ERROR_NONE;
@@ -522,7 +649,7 @@ server_ptrace_insert_breakpoint (ServerHandle *handle, guint64 address, guint32
 	breakpoint->id = mono_debugger_breakpoint_manager_get_next_id ();
 	breakpoint->dr_index = -1;
 
-	result = do_enable (handle, breakpoint);
+	result = x86_arch_enable_breakpoint (handle, breakpoint);
 	if (result != COMMAND_ERROR_NONE) {
 		mono_debugger_breakpoint_manager_unlock ();
 		g_free (breakpoint);
@@ -556,7 +683,7 @@ server_ptrace_remove_breakpoint (ServerHandle *handle, guint32 bhandle)
 		goto out;
 	}
 
-	result = do_disable (handle, breakpoint);
+	result = x86_arch_disable_breakpoint (handle, breakpoint);
 	if (result != COMMAND_ERROR_NONE)
 		goto out;
 
@@ -568,26 +695,6 @@ server_ptrace_remove_breakpoint (ServerHandle *handle, guint32 bhandle)
 	return result;
 }
 
-static ServerCommandError
-server_ptrace_init_after_fork (ServerHandle *handle)
-{
-	GPtrArray *breakpoints;
-	int i;
-
-	mono_debugger_breakpoint_manager_lock ();
-
-	breakpoints = mono_debugger_breakpoint_manager_get_breakpoints (handle->bpm);
-	for (i = 0; i < breakpoints->len; i++) {
-		BreakpointInfo *info = g_ptr_array_index (breakpoints, i);
-
-		if (info->dr_index >= 0)
-			do_disable (handle, info);
-	}
-
-	mono_debugger_breakpoint_manager_unlock ();
-	return COMMAND_ERROR_NONE;
-}
-
 static ServerCommandError
 find_free_hw_register (ServerHandle *handle, guint32 *idx)
 {
@@ -631,7 +738,7 @@ server_ptrace_insert_hw_breakpoint (ServerHandle *handle, guint32 type, guint32
 	breakpoint->is_hardware_bpt = TRUE;
 	breakpoint->dr_index = *idx;
 
-	result = do_enable (handle, breakpoint);
+	result = x86_arch_enable_breakpoint (handle, breakpoint);
 	if (result != COMMAND_ERROR_NONE) {
 		mono_debugger_breakpoint_manager_unlock ();
 		g_free (breakpoint);
@@ -660,7 +767,7 @@ server_ptrace_enable_breakpoint (ServerHandle *handle, guint32 bhandle)
 		return COMMAND_ERROR_NO_SUCH_BREAKPOINT;
 	}
 
-	result = do_enable (handle, breakpoint);
+	result = x86_arch_enable_breakpoint (handle, breakpoint);
 	breakpoint->enabled = TRUE;
 	mono_debugger_breakpoint_manager_unlock ();
 	return result;
@@ -679,7 +786,7 @@ server_ptrace_disable_breakpoint (ServerHandle *handle, guint32 bhandle)
 		return COMMAND_ERROR_NO_SUCH_BREAKPOINT;
 	}
 
-	result = do_disable (handle, breakpoint);
+	result = x86_arch_disable_breakpoint (handle, breakpoint);
 	breakpoint->enabled = FALSE;
 	mono_debugger_breakpoint_manager_unlock ();
 	return result;
@@ -728,6 +835,7 @@ server_ptrace_call_method (ServerHandle *handle, guint64 method_address,
 {
 	ServerCommandError result = COMMAND_ERROR_NONE;
 	ArchInfo *arch = handle->arch;
+	CallbackData *cdata;
 	long new_rsp;
 
 	guint8 code[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
@@ -735,8 +843,7 @@ server_ptrace_call_method (ServerHandle *handle, guint64 method_address,
 			  0xcc, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
 	int size = sizeof (code);
 
-	if (arch->saved_regs)
-		return COMMAND_ERROR_RECURSIVE_CALL;
+	cdata = g_new0 (CallbackData, 1);
 
 	new_rsp = INFERIOR_REG_RSP (arch->current_regs) - AMD64_RED_ZONE_SIZE - size - 16;
 	new_rsp &= 0xfffffffffffffff0L;
@@ -744,11 +851,12 @@ server_ptrace_call_method (ServerHandle *handle, guint64 method_address,
 	*((guint64 *) code) = new_rsp + 16;
 	*((guint64 *) (code+8)) = callback_argument;
 
-	arch->saved_regs = g_memdup (&arch->current_regs, sizeof (arch->current_regs));
-	arch->saved_fpregs = g_memdup (&arch->current_fpregs, sizeof (arch->current_fpregs));
-	arch->call_address = new_rsp + 16;
-	arch->callback_argument = callback_argument;
-	arch->saved_signal = handle->inferior->last_signal;
+	memcpy (&cdata->saved_regs, &arch->current_regs, sizeof (arch->current_regs));
+	memcpy (&cdata->saved_fpregs, &arch->current_fpregs, sizeof (arch->current_fpregs));
+	cdata->call_address = new_rsp + 16;
+	cdata->stack_pointer = new_rsp + 8;
+	cdata->callback_argument = callback_argument;
+	cdata->saved_signal = handle->inferior->last_signal;
 	handle->inferior->last_signal = 0;
 
 	server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
@@ -760,6 +868,8 @@ server_ptrace_call_method (ServerHandle *handle, guint64 method_address,
 	INFERIOR_REG_RSI (arch->current_regs) = method_argument2;
 	INFERIOR_REG_RSP (arch->current_regs) = new_rsp;
 
+	g_ptr_array_add (arch->callback_stack, cdata);
+
 	result = _server_ptrace_set_registers (handle->inferior, &arch->current_regs);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
@@ -774,11 +884,13 @@ server_ptrace_call_method (ServerHandle *handle, guint64 method_address,
 
 static ServerCommandError
 server_ptrace_call_method_1 (ServerHandle *handle, guint64 method_address,
-			     guint64 method_argument, const gchar *string_argument,
+			     guint64 method_argument, guint64 data_argument,
+			     guint64 data_argument2, const gchar *string_argument,
 			     guint64 callback_argument)
 {
 	ServerCommandError result = COMMAND_ERROR_NONE;
 	ArchInfo *arch = handle->arch;
+	CallbackData *cdata;
 	long new_rsp;
 
 	static guint8 static_code[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
@@ -790,8 +902,7 @@ server_ptrace_call_method_1 (ServerHandle *handle, guint64 method_address,
 	memcpy (code, static_code, static_size);
 	strcpy (code + static_size, string_argument);
 
-	if (arch->saved_regs)
-		return COMMAND_ERROR_RECURSIVE_CALL;
+	cdata = g_new0 (CallbackData, 1);
 
 	new_rsp = INFERIOR_REG_RSP (arch->current_regs) - AMD64_RED_ZONE_SIZE - size - 16;
 	new_rsp &= 0xfffffffffffffff0L;
@@ -799,11 +910,12 @@ server_ptrace_call_method_1 (ServerHandle *handle, guint64 method_address,
 	*((guint64 *) code) = new_rsp + 16;
 	*((guint64 *) (code+8)) = callback_argument;
 
-	arch->saved_regs = g_memdup (&arch->current_regs, sizeof (arch->current_regs));
-	arch->saved_fpregs = g_memdup (&arch->current_fpregs, sizeof (arch->current_fpregs));
-	arch->call_address = new_rsp + 16;
-	arch->callback_argument = callback_argument;
-	arch->saved_signal = handle->inferior->last_signal;
+	memcpy (&cdata->saved_regs, &arch->current_regs, sizeof (arch->current_regs));
+	memcpy (&cdata->saved_fpregs, &arch->current_fpregs, sizeof (arch->current_fpregs));
+	cdata->call_address = new_rsp + 16;
+	cdata->stack_pointer = new_rsp + 8;
+	cdata->callback_argument = callback_argument;
+	cdata->saved_signal = handle->inferior->last_signal;
 	handle->inferior->last_signal = 0;
 
 	server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
@@ -812,9 +924,13 @@ server_ptrace_call_method_1 (ServerHandle *handle, guint64 method_address,
 
 	INFERIOR_REG_RIP (arch->current_regs) = method_address;
 	INFERIOR_REG_RDI (arch->current_regs) = method_argument;
-	INFERIOR_REG_RSI (arch->current_regs) = new_rsp + static_size;
+	INFERIOR_REG_RSI (arch->current_regs) = data_argument;
+	INFERIOR_REG_RDX (arch->current_regs) = data_argument2;
+	INFERIOR_REG_RCX (arch->current_regs) = new_rsp + static_size;
 	INFERIOR_REG_RSP (arch->current_regs) = new_rsp;
 
+	g_ptr_array_add (arch->callback_stack, cdata);
+
 	result = _server_ptrace_set_registers (handle->inferior, &arch->current_regs);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
@@ -824,19 +940,17 @@ server_ptrace_call_method_1 (ServerHandle *handle, guint64 method_address,
 
 static ServerCommandError
 server_ptrace_call_method_2 (ServerHandle *handle, guint64 method_address,
-			     guint64 method_argument, guint64 callback_argument)
+			     guint32 data_size, gconstpointer data_buffer,
+			     guint64 callback_argument)
 {
 	ServerCommandError result = COMMAND_ERROR_NONE;
 	ArchInfo *arch = handle->arch;
-	RuntimeInvokeData *rdata;
+	CallbackData *cdata;
 	long new_rsp;
 
-	int size = 116;
+	int size = 113 + data_size;
 	guint8 *code = g_malloc0 (size);
 
-	if (arch->saved_regs)
-		return COMMAND_ERROR_RECURSIVE_CALL;
-
 	new_rsp = INFERIOR_REG_RSP (arch->current_regs) - AMD64_RED_ZONE_SIZE - size - 16;
 	new_rsp &= 0xfffffffffffffff0L;
 
@@ -854,17 +968,25 @@ server_ptrace_call_method_2 (ServerHandle *handle, guint64 method_address,
 	*((guint64 *) (code+88)) = INFERIOR_REG_R13 (arch->current_regs);
 	*((guint64 *) (code+96)) = INFERIOR_REG_R14 (arch->current_regs);
 	*((guint64 *) (code+104)) = INFERIOR_REG_R15 (arch->current_regs);
-	*((guint8 *) (code+112)) = 0xcc;
+	*((guint8 *) (code+data_size+112)) = 0xcc;
 
-	rdata = g_new0 (RuntimeInvokeData, 1);
-	rdata->saved_regs = g_memdup (&arch->current_regs, sizeof (arch->current_regs));
-	rdata->saved_fpregs = g_memdup (&arch->current_fpregs, sizeof (arch->current_fpregs));
-	rdata->call_address = new_rsp + size - 1;
-	rdata->exc_address = 0;
-	rdata->callback_argument = callback_argument;
-	rdata->saved_signal = handle->inferior->last_signal;
+	cdata = g_new0 (CallbackData, 1);
+	memcpy (&cdata->saved_regs, &arch->current_regs, sizeof (arch->current_regs));
+	memcpy (&cdata->saved_fpregs, &arch->current_fpregs, sizeof (arch->current_fpregs));
+	cdata->call_address = new_rsp + size - 1;
+	cdata->stack_pointer = new_rsp + 8;
+	cdata->exc_address = 0;
+	cdata->callback_argument = callback_argument;
+	cdata->saved_signal = handle->inferior->last_signal;
+	cdata->pushed_registers = new_rsp + 8;
 	handle->inferior->last_signal = 0;
 
+	if (data_size > 0) {
+		memcpy (code+112, data_buffer, data_size);
+		cdata->data_pointer = new_rsp + 112;
+		cdata->data_size = data_size;
+	}
+
 	server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
 	g_free (code);
 	if (result != COMMAND_ERROR_NONE)
@@ -872,10 +994,10 @@ server_ptrace_call_method_2 (ServerHandle *handle, guint64 method_address,
 
 	INFERIOR_REG_RIP (arch->current_regs) = method_address;
 	INFERIOR_REG_RDI (arch->current_regs) = new_rsp + 8;
-	INFERIOR_REG_RSI (arch->current_regs) = method_argument;
+	INFERIOR_REG_RSI (arch->current_regs) = new_rsp + 112;
 	INFERIOR_REG_RSP (arch->current_regs) = new_rsp;
 
-	g_ptr_array_add (arch->rti_stack, rdata);
+	g_ptr_array_add (arch->callback_stack, cdata);
 
 	result = _server_ptrace_set_registers (handle->inferior, &arch->current_regs);
 	if (result != COMMAND_ERROR_NONE)
@@ -893,7 +1015,7 @@ server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 {
 	ServerCommandError result = COMMAND_ERROR_NONE;
 	ArchInfo *arch = handle->arch;
-	RuntimeInvokeData *rdata;
+	CallbackData *cdata;
 	guint64 new_rsp;
 	int i;
 
@@ -909,9 +1031,6 @@ server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 	memcpy (code, static_code, static_size);
 	memcpy (code + static_size, blob_data, blob_size);
 
-	if (arch->saved_regs)
-		return COMMAND_ERROR_RECURSIVE_CALL;
-
 	new_rsp = INFERIOR_REG_RSP (arch->current_regs) - AMD64_RED_ZONE_SIZE - size - 16;
 	new_rsp &= 0xfffffffffffffff0L;
 
@@ -927,14 +1046,15 @@ server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 	*((guint64 *) code) = new_rsp + static_size - 1;
 	*((guint64 *) (code+8)) = callback_argument;
 
-	rdata = g_new0 (RuntimeInvokeData, 1);
-	rdata->saved_regs = g_memdup (&arch->current_regs, sizeof (arch->current_regs));
-	rdata->saved_fpregs = g_memdup (&arch->current_fpregs, sizeof (arch->current_fpregs));
-	rdata->call_address = new_rsp + static_size - 1;
-	rdata->exc_address = new_rsp + 16;
-	rdata->callback_argument = callback_argument;
-	rdata->debug = debug;
-	rdata->saved_signal = handle->inferior->last_signal;
+	cdata = g_new0 (CallbackData, 1);
+	memcpy (&cdata->saved_regs, &arch->current_regs, sizeof (arch->current_regs));
+	memcpy (&cdata->saved_fpregs, &arch->current_fpregs, sizeof (arch->current_fpregs));
+	cdata->call_address = new_rsp + static_size - 1;
+	cdata->stack_pointer = new_rsp + 8;
+	cdata->exc_address = new_rsp + 16;
+	cdata->callback_argument = callback_argument;
+	cdata->debug = debug;
+	cdata->saved_signal = handle->inferior->last_signal;
 	handle->inferior->last_signal = 0;
 
 	server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
@@ -949,7 +1069,7 @@ server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 	INFERIOR_REG_RCX (arch->current_regs) = new_rsp + 16;
 	INFERIOR_REG_RSP (arch->current_regs) = new_rsp;
 
-	g_ptr_array_add (arch->rti_stack, rdata);
+	g_ptr_array_add (arch->callback_stack, cdata);
 
 	result = _server_ptrace_set_registers (handle->inferior, &arch->current_regs);
 	if (result != COMMAND_ERROR_NONE)
@@ -958,28 +1078,170 @@ server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 	return server_ptrace_continue (handle);
 }
 
-static ServerCommandError
-server_ptrace_abort_invoke (ServerHandle *handle)
+static int
+find_code_buffer_slot (MonoRuntimeInfo *runtime)
 {
-	RuntimeInvokeData *rdata;
+	int i;
 
-	rdata = get_runtime_invoke_data (handle->arch);
-	if (!rdata)
-		return COMMAND_ERROR_UNKNOWN_ERROR;
+	for (i = 0; i < runtime->executable_code_total_chunks; i++) {
+		if (runtime->executable_code_bitfield [i])
+			continue;
 
-	if (_server_ptrace_set_registers (handle->inferior, rdata->saved_regs) != COMMAND_ERROR_NONE)
+		runtime->executable_code_bitfield [i] = 1;
+		return i;
+	}
+
+	return -1;
+}
+
+static ServerCommandError
+server_ptrace_execute_instruction (ServerHandle *handle, const guint8 *instruction,
+				   guint32 size, gboolean update_ip)
+{
+	MonoRuntimeInfo *runtime;
+	ServerCommandError result;
+	CodeBufferData *data;
+	guint64 code_address;
+	int slot;
+
+	runtime = handle->mono_runtime;
+	g_assert (runtime);
+
+	if (!runtime->executable_code_buffer)
+		return COMMAND_ERROR_INTERNAL_ERROR;
+
+	slot = find_code_buffer_slot (runtime);
+	if (slot < 0)
+		return COMMAND_ERROR_INTERNAL_ERROR;
+
+	if (size > runtime->executable_code_chunk_size)
+		return COMMAND_ERROR_INTERNAL_ERROR;
+	if (handle->arch->code_buffer)
+		return COMMAND_ERROR_INTERNAL_ERROR;
+
+	code_address = runtime->executable_code_buffer + slot * runtime->executable_code_chunk_size;
+
+	data = g_new0 (CodeBufferData, 1);
+	data->slot = slot;
+	data->insn_size = size;
+	data->update_ip = update_ip;
+	data->original_rip = INFERIOR_REG_RIP (handle->arch->current_regs);
+	data->code_address = code_address;
+
+	handle->arch->code_buffer = data;
+
+	result = server_ptrace_write_memory (handle, code_address, size, instruction);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	INFERIOR_REG_RIP (handle->arch->current_regs) = code_address;
+
+	result = _server_ptrace_set_registers (handle->inferior, &handle->arch->current_regs);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	return server_ptrace_step (handle);
+}
+
+static ServerCommandError
+server_ptrace_mark_rti_frame (ServerHandle *handle)
+{
+	CallbackData *cdata;
+
+	cdata = get_callback_data (handle->arch);
+	if (!cdata)
+		return COMMAND_ERROR_NO_CALLBACK_FRAME;
+
+	cdata->rti_frame = INFERIOR_REG_RSP (handle->arch->current_regs) + 8;
+	return COMMAND_ERROR_NONE;
+}
+
+static ServerCommandError
+server_ptrace_abort_invoke (ServerHandle *handle, guint64 stack_pointer)
+{
+	CallbackData *cdata;
+
+	cdata = get_callback_data (handle->arch);
+	if (!cdata)
+		return COMMAND_ERROR_NO_CALLBACK_FRAME;
+
+	if (cdata->rti_frame && (stack_pointer < cdata->rti_frame))
+		return COMMAND_ERROR_NO_CALLBACK_FRAME;
+	if (stack_pointer < cdata->stack_pointer)
+		return COMMAND_ERROR_NO_CALLBACK_FRAME;
+
+	if (_server_ptrace_set_registers (handle->inferior, &cdata->saved_regs) != COMMAND_ERROR_NONE)
 		g_error (G_STRLOC ": Can't restore registers after returning from a call");
 
-	if (_server_ptrace_set_fp_registers (handle->inferior, rdata->saved_fpregs) != COMMAND_ERROR_NONE)
+	if (_server_ptrace_set_fp_registers (handle->inferior, &cdata->saved_fpregs) != COMMAND_ERROR_NONE)
 		g_error (G_STRLOC ": Can't restore FP registers after returning from a call");
 
-	handle->inferior->last_signal = rdata->saved_signal;
-	g_free (rdata->saved_regs);
-	g_free (rdata->saved_fpregs);
-	g_ptr_array_remove (handle->arch->rti_stack, rdata);
+	handle->inferior->last_signal = cdata->saved_signal;
+	g_ptr_array_remove (handle->arch->callback_stack, cdata);
 
 	x86_arch_get_registers (handle);
-	g_free (rdata);
+	g_free (cdata);
 
 	return COMMAND_ERROR_NONE;
 }
+
+static ServerCommandError
+server_ptrace_get_callback_frame (ServerHandle *handle, guint64 stack_pointer,
+				  gboolean exact_match, guint64 *registers)
+{
+	int i;
+
+	for (i = 0; i < handle->arch->callback_stack->len; i++) {
+		CallbackData *cdata = g_ptr_array_index (handle->arch->callback_stack, i);
+
+		if (cdata->rti_frame) {
+			if (exact_match) {
+				if (cdata->rti_frame != stack_pointer)
+					continue;
+			} else {
+				if (cdata->rti_frame < stack_pointer)
+					continue;
+			}
+		} else {
+			if (exact_match) {
+				if (cdata->stack_pointer != stack_pointer)
+					continue;
+			} else {
+				if (cdata->stack_pointer < stack_pointer)
+					continue;
+			}
+		}
+
+		registers [DEBUGGER_REG_R15] = (guint64) INFERIOR_REG_R15 (cdata->saved_regs);
+		registers [DEBUGGER_REG_R14] = (guint64) INFERIOR_REG_R14 (cdata->saved_regs);
+		registers [DEBUGGER_REG_R13] = (guint64) INFERIOR_REG_R13 (cdata->saved_regs);
+		registers [DEBUGGER_REG_R12] = (guint64) INFERIOR_REG_R12 (cdata->saved_regs);
+		registers [DEBUGGER_REG_RBP] = (guint64) INFERIOR_REG_RBP (cdata->saved_regs);
+		registers [DEBUGGER_REG_RBX] = (guint64) INFERIOR_REG_RBX (cdata->saved_regs);
+		registers [DEBUGGER_REG_R11] = (guint64) INFERIOR_REG_R11 (cdata->saved_regs);
+		registers [DEBUGGER_REG_R10] = (guint64) INFERIOR_REG_R10 (cdata->saved_regs);
+		registers [DEBUGGER_REG_R9] = (guint64) INFERIOR_REG_R9 (cdata->saved_regs);
+		registers [DEBUGGER_REG_R8] = (guint64) INFERIOR_REG_R8 (cdata->saved_regs);
+		registers [DEBUGGER_REG_RAX] = (guint64) INFERIOR_REG_RAX (cdata->saved_regs);
+		registers [DEBUGGER_REG_RCX] = (guint64) INFERIOR_REG_RCX (cdata->saved_regs);
+		registers [DEBUGGER_REG_RDX] = (guint64) INFERIOR_REG_RDX (cdata->saved_regs);
+		registers [DEBUGGER_REG_RSI] = (guint64) INFERIOR_REG_RSI (cdata->saved_regs);
+		registers [DEBUGGER_REG_RDI] = (guint64) INFERIOR_REG_RDI (cdata->saved_regs);
+		registers [DEBUGGER_REG_ORIG_RAX] = (guint64) INFERIOR_REG_ORIG_RAX (cdata->saved_regs);
+		registers [DEBUGGER_REG_RIP] = (guint64) INFERIOR_REG_RIP (cdata->saved_regs);
+		registers [DEBUGGER_REG_CS] = (guint64) INFERIOR_REG_CS (cdata->saved_regs);
+		registers [DEBUGGER_REG_EFLAGS] = (guint64) INFERIOR_REG_EFLAGS (cdata->saved_regs);
+		registers [DEBUGGER_REG_RSP] = (guint64) INFERIOR_REG_RSP (cdata->saved_regs);
+		registers [DEBUGGER_REG_SS] = (guint64) INFERIOR_REG_SS (cdata->saved_regs);
+		registers [DEBUGGER_REG_FS_BASE] = (guint64) INFERIOR_REG_FS_BASE (cdata->saved_regs);
+		registers [DEBUGGER_REG_GS_BASE] = (guint64) INFERIOR_REG_GS_BASE (cdata->saved_regs);
+		registers [DEBUGGER_REG_DS] = (guint64) INFERIOR_REG_DS (cdata->saved_regs);
+		registers [DEBUGGER_REG_ES] = (guint64) INFERIOR_REG_ES (cdata->saved_regs);
+		registers [DEBUGGER_REG_FS] = (guint64) INFERIOR_REG_FS (cdata->saved_regs);
+		registers [DEBUGGER_REG_GS] = (guint64) INFERIOR_REG_GS (cdata->saved_regs);
+
+		return COMMAND_ERROR_NONE;
+	}
+
+	return COMMAND_ERROR_NO_CALLBACK_FRAME;
+}
