@@ -1,5 +1,5 @@
 using System;
-using Mono.Debugger.Backends;
+using Mono.Debugger.Backend;
 using System.Runtime.Serialization;
 using System.Xml;
 
@@ -17,6 +17,12 @@ namespace Mono.Debugger
 	// </summary>
 	public abstract class Breakpoint : Event
 	{
+		internal abstract BreakpointHandle Resolve (Thread target, StackFrame frame);
+
+		public virtual bool HideFromUser {
+			get { return false; }
+		}
+
 		public override void Remove (Thread target)
 		{
 			Deactivate (target);
