@@ -34,30 +34,27 @@ namespace Mono.Debugger.Frontend
 				throw new ScriptingException ("Cannot find source file `{0}'.",
 							      filename);
 
-			SourceMethod source = file.FindMethod (line);
+			MethodSource source = file.FindMethod (line);
 			if (source == null)
 				throw new ScriptingException (
 					"Cannot find method corresponding to line {0} in `{1}'.",
 					line, file.Name);
 
-			return new SourceLocation (source, line);
+			return new SourceLocation (source, file, line);
 		}
 
-		protected SourceLocation DoParseExpression (Thread target, LocationType type, string arg)
+		protected SourceLocation DoParseExpression (Thread target, StackFrame frame,
+							    LocationType type, string arg)
 		{
 			ScriptingContext context = new ScriptingContext (Interpreter);
-			context.CurrentFrame = target.CurrentFrame;
+			context.CurrentThread = frame.Thread;
+			context.CurrentFrame = frame;
 
 			Expression expr = Interpreter.ExpressionParser.Parse (arg);
 			if (expr == null)
 				throw new ScriptingException ("Cannot resolve expression `{0}'.", arg);
 
-			MethodExpression mexpr;
-			try {
-				mexpr = expr.ResolveMethod (context, type);
-			} catch {
-				return null;
-			}
+			MethodExpression mexpr = expr.ResolveMethod (context, type);
 
 			if (mexpr != null)
 				return mexpr.EvaluateSource (context);
@@ -65,7 +62,8 @@ namespace Mono.Debugger.Frontend
 				return context.FindMethod (arg);
 		}
 
-		public static bool ParseLocation (Thread target, string arg, out SourceLocation location)
+		public static bool ParseLocation (Thread target, StackFrame frame,
+						  string arg, out SourceLocation location)
 		{
 			int line;
 			int pos = arg.IndexOf (':');
@@ -77,7 +75,7 @@ namespace Mono.Debugger.Frontend
 					throw new ScriptingException ("Expected filename:line");
 				}
 
-				location = FindFile (target, filename, line);
+				location = FindFile (frame.Thread, filename, line);
 				return true;
 			}
 
@@ -88,41 +86,36 @@ namespace Mono.Debugger.Frontend
 				return false;
 			}
 
-			StackFrame frame = target.CurrentFrame;
-			if ((frame.SourceAddress == null) || (frame.SourceAddress.Location == null))
+			if ((frame == null) || (frame.SourceLocation == null) ||
+			    (frame.SourceLocation.FileName == null))
 				throw new ScriptingException (
 					"Current stack frame doesn't have source code");
 
-			location = frame.SourceAddress.Location;
-			if (location.FileName == null)
-				throw new ScriptingException (
-					"Current stack frame doesn't have source code");
-
-			location = FindFile (target, location.FileName, line);
+			location = FindFile (frame.Thread, frame.SourceLocation.FileName, line);
 			return true;
 		}
 
-		protected SourceLocation DoParse (Thread target, LocationType type, string arg)
+		protected SourceLocation DoParse (Thread target, StackFrame frame,
+						  LocationType type, string arg)
 		{
 			if (type != LocationType.Default)
-				return DoParseExpression (target, type, arg);
+				return DoParseExpression (target, frame, type, arg);
 
 			SourceLocation location;
-			if (ParseLocation (target, arg, out location))
+			if (ParseLocation (target, frame, arg, out location))
 				return location;
 
-			return DoParseExpression (target, type, arg);
+			return DoParseExpression (target, frame, type, arg);
 		}
 
-		public SourceLocation Parse (Thread target, LocationType type, string arg)
+		public SourceLocation Parse (Thread target, StackFrame frame,
+					     LocationType type, string arg)
 		{
 			try {
-				return DoParse (target, type, arg);
+				return DoParse (target, frame, type, arg);
 			} catch (ScriptingException ex) {
 				throw new TargetException (TargetError.LocationInvalid, ex.Message);
 			}
-
-			return null;
 		}
 
 		public string EvaluateExpression (ScriptingContext context, string text,
@@ -665,7 +658,7 @@ namespace Mono.Debugger.Frontend
 					"Keyword `this' not allowed: current method is " +
 					"either static or unmanaged.");
 
-			var = method.This;
+			var = method.GetThis (context.CurrentThread);
 			resolved = true;
 			return this;
 		}
@@ -867,19 +860,53 @@ namespace Mono.Debugger.Frontend
                         return String.Concat (nsn, ".", name);
                 }
 
+		TargetVariable GetVariableByName (StackFrame frame, string name)
+		{
+			TargetVariable[] locals = frame.Method.GetLocalVariables (frame.Thread);
+			foreach (TargetVariable var in locals) {
+				if ((var.Name == name) && var.IsInScope (frame.TargetAddress))
+					return var;
+			}
+
+			TargetVariable[] param_vars = frame.Method.GetParameters (frame.Thread);
+			foreach (TargetVariable var in param_vars) {
+				if ((var.Name == name) && var.IsInScope (frame.TargetAddress))
+					return var;
+			}
+
+			return null;
+		}
+
 		MemberExpression LookupMember (ScriptingContext context, StackFrame frame,
 					       string full_name)
 		{
+			MemberExpression member;
+
+			TargetFunctionType function = frame.Function;
+			if (function != null) {
+				member = StructAccessExpression.FindMember (
+					frame.Thread, function.DeclaringType, null,
+					full_name, true, true);
+				if (member != null)
+					return member;
+			}
+
 			Method method = frame.Method;
-			if ((method == null) || (method.DeclaringType == null))
+			if (method == null)
+				return null;
+
+			TargetClassType decl_type = method.GetDeclaringType (context.CurrentThread);
+			if (decl_type == null)
 				return null;
 
 			TargetClassObject instance = null;
-			if (method.HasThis)
-				instance = (TargetClassObject) method.This.GetObject (frame);
+			if (method.HasThis) {
+				TargetVariable this_var = method.GetThis (context.CurrentThread);
+				instance = (TargetClassObject) this_var.GetObject (frame);
+			}
 
-			MemberExpression member = StructAccessExpression.FindMember (
-				frame.Thread, method.DeclaringType, instance, full_name, true, true);
+			member = StructAccessExpression.FindMember (
+				context.CurrentThread, decl_type, instance, full_name, true, true);
 			if (member == null)
 				return null;
 
@@ -908,25 +935,26 @@ namespace Mono.Debugger.Frontend
 
 		protected override Expression DoResolve (ScriptingContext context)
 		{
-			StackFrame frame = context.CurrentFrame;
-			if (frame.Method != null) {
-				TargetVariable var = frame.Method.GetVariableByName (
-					frame.TargetAddress, name);
-				if (var != null)
-					return new VariableAccessExpression (var);
-			}
+			if (context.HasFrame) {
+				StackFrame frame = context.CurrentFrame;
+				if ((frame.Method != null) && frame.Method.IsLoaded) {
+					TargetVariable var = GetVariableByName (frame, name);
+					if (var != null)
+						return new VariableAccessExpression (var);
+				}
 
-			Expression expr = Lookup (context, frame);
-			if (expr != null)
-				return expr;
+				Expression expr = Lookup (context, frame);
+				if (expr != null)
+					return expr;
+			}
 
 			SourceLocation location = context.FindMethod (name);
 			if (location != null)
 				return new SourceExpression (location);
 
-			expr = DoResolveType (context);
-			if (expr != null)
-				return expr;
+			Expression texpr = DoResolveType (context);
+			if (texpr != null)
+				return texpr;
 
 			throw new ScriptingException ("No symbol `{0}' in current context.", Name);
 		}
@@ -934,9 +962,7 @@ namespace Mono.Debugger.Frontend
 		protected override MethodExpression DoResolveMethod (ScriptingContext context,
 								     LocationType type)
 		{
-			StackFrame frame = context.CurrentFrame;
-
-			MemberExpression member;
+			MemberExpression member = null;
 			if (type == LocationType.Constructor) {
 				Expression texpr = ResolveType (context);
 				if (texpr == null)
@@ -947,9 +973,9 @@ namespace Mono.Debugger.Frontend
 					return null;
 
 				member = StructAccessExpression.FindMember (
-					frame.Thread, ctype, null, ".ctor", false, true);
-			} else {
-				member = Lookup (context, frame);
+					context.CurrentThread, ctype, null, ".ctor", false, true);
+			} else if (context.HasFrame) {
+				member = Lookup (context, context.CurrentFrame);
 			}
 
 			if (member != null)
@@ -1477,7 +1503,7 @@ namespace Mono.Debugger.Frontend
 				return (long) val;
 			else if (val is TargetPointerObject) {
 				TargetPointerObject pobj = (TargetPointerObject) val;
-				return pobj.Address.Address;
+				return pobj.GetAddress (context.CurrentThread).Address;
 			} else
 				throw new ScriptingException ("Cannot evaluate expression `{0}'", expr);
 		}
@@ -1591,7 +1617,7 @@ namespace Mono.Debugger.Frontend
 		public override TargetAddress EvaluateAddress (ScriptingContext context)
 		{
 			TargetPointerObject pobj = (TargetPointerObject) EvaluateObject (context);
-			return pobj.Address;
+			return pobj.GetAddress (context.CurrentThread);
 		}
 
 		protected override bool DoAssign (ScriptingContext context, TargetObject tobj)
@@ -1619,6 +1645,7 @@ namespace Mono.Debugger.Frontend
 		public readonly TargetClassType Type;
 		public readonly TargetMemberInfo Member;
 		protected readonly TargetClassObject instance;
+		TargetClass class_info;
 
 		protected StructAccessExpression (TargetClassType type,
 						  TargetClassObject instance,
@@ -1659,6 +1686,17 @@ namespace Mono.Debugger.Frontend
 				!func.IsStatic, func.IsStatic);
 		}
 
+		protected void ResolveClass (Thread target)
+		{
+			if (class_info != null)
+				return;
+
+			class_info = Type.GetClass (target);
+			if (class_info == null)
+				throw new ScriptingException ("Class `{0}' not initialized yet.",
+							      Type.Name);
+		}
+
 		protected override MethodExpression DoResolveMethod (ScriptingContext context,
 								     LocationType type)
 		{
@@ -1707,12 +1745,8 @@ namespace Mono.Debugger.Frontend
 
 		protected TargetObject GetField (Thread target, TargetFieldInfo field)
 		{
-			if (field.HasConstValue)
-				return field.Type.Language.CreateInstance (target, field.ConstValue);
-			else if (field.IsStatic)
-				return Type.GetStaticField (target, field);
-			else
-				return InstanceObject.GetField (target, field);
+			ResolveClass (target);
+			return class_info.GetField (target, InstanceObject, field);
 		}
 
 		protected TargetObject GetProperty (ScriptingContext context,
@@ -1742,50 +1776,53 @@ namespace Mono.Debugger.Frontend
 							      Name, member.GetType ());
 		}
 
-		public static MemberExpression FindMember (Thread target, TargetClassType stype,
+		public static MemberExpression FindMember (Thread target, TargetStructType stype,
 							   TargetClassObject instance, string name,
 							   bool search_static, bool search_instance)
 		{
 		again:
-			TargetMemberInfo member = stype.FindMember (
+			TargetClassType ctype = stype as TargetClassType;
+			if (ctype == null)
+				return null;
+
+			TargetMemberInfo member = ctype.FindMember (
 				name, search_static, search_instance);
 
 			if (member != null)
-				return new StructAccessExpression (stype, instance, member);
+				return new StructAccessExpression (ctype, instance, member);
 
 			ArrayList methods = new ArrayList ();
 			bool is_instance = false;
 			bool is_static = false;
 
 			if (name == ".ctor") {
-				foreach (TargetMethodInfo method in stype.Constructors) {
+				foreach (TargetMethodInfo method in ctype.Constructors) {
+					if (method.IsStatic)
+						continue;
 					methods.Add (method.Type);
 					is_instance = true;
 				}
 			} else if (name == ".cctor") {
-				foreach (TargetMethodInfo method in stype.StaticConstructors) {
+				foreach (TargetMethodInfo method in ctype.Constructors) {
+					if (!method.IsStatic)
+						continue;
 					methods.Add (method.Type);
 					is_static = true;
 				}
 			} else {
-				if (search_instance) {
-					foreach (TargetMethodInfo method in stype.Methods) {
-						if (method.Name != name)
-							continue;
+				foreach (TargetMethodInfo method in ctype.Methods) {
+					if (method.IsStatic && !search_static)
+						continue;
+					if (!method.IsStatic && !search_instance)
+						continue;
+					if (method.Name != name)
+						continue;
 
-						methods.Add (method.Type);
-						is_instance = true;
-					}
-				}
-
-				if (search_static) {
-					foreach (TargetMethodInfo method in stype.StaticMethods) {
-						if (method.Name != name)
-							continue;
-
-						methods.Add (method.Type);
+					methods.Add (method.Type);
+					if (method.IsStatic)
 						is_static = true;
-					}
+					else
+						is_instance = true;
 				}
 			}
 
@@ -1793,12 +1830,16 @@ namespace Mono.Debugger.Frontend
 				TargetFunctionType[] funcs = new TargetFunctionType [methods.Count];
 				methods.CopyTo (funcs, 0);
 				return new MethodGroupExpression (
-					stype, instance, name, funcs, is_instance, is_static);
+					ctype, instance, name, funcs, is_instance, is_static);
 			}
 
-			TargetClassType ctype = stype as TargetClassType;
-			if ((ctype != null) && ctype.HasParent) {
-				stype = ctype.ParentType;
+			if (ctype.HasParent) {
+				stype = ctype.GetParentType (target);
+				if (instance != null) {
+					instance = instance.GetParentObject (target);
+					if (instance == null)
+						return null;
+				}
 				goto again;
 			}
 
@@ -1886,10 +1927,8 @@ namespace Mono.Debugger.Frontend
 
 		protected void SetField (Thread target, TargetFieldInfo field, TargetObject obj)
 		{
-			if (field.IsStatic)
-				Type.SetStaticField (target, field, obj);
-			else
-				InstanceObject.SetField (target, field, obj);
+			ResolveClass (target);
+			class_info.SetField (target, InstanceObject, field, obj);
 		}
 
 		protected override bool DoAssign (ScriptingContext context, TargetObject obj)
@@ -2009,7 +2048,7 @@ namespace Mono.Debugger.Frontend
 				throw new ScriptingException (
 					"Expression `{0}' is not a pointer type.", expr.Name);
 
-			return pobj.Address;
+			return pobj.GetAddress (context.CurrentThread);
 		}
 	}
 
@@ -2070,7 +2109,7 @@ namespace Mono.Debugger.Frontend
 				throw new ScriptingException (
 					"Cannot take address of expression `{0}'", expr.Name);
 
-			return obj.Address;
+			return obj.GetAddress (context.CurrentThread);
 		}
 	}
 
@@ -2314,10 +2353,10 @@ namespace Mono.Debugger.Frontend
 			return this;
 		}
 
-		static TargetClassObject TryParentCast (ScriptingContext context, Thread target,
+		static TargetClassObject TryParentCast (ScriptingContext context,
 							TargetClassObject source,
-							TargetClassType source_type,
-							TargetClassType target_type)
+							TargetStructType source_type,
+							TargetStructType target_type)
 		{
 			if (source_type == target_type)
 				return source;
@@ -2325,23 +2364,23 @@ namespace Mono.Debugger.Frontend
 			if (!source_type.HasParent)
 				return null;
 
-			source = TryParentCast (
-				context, target, source, source_type.ParentType, target_type);
+			TargetStructType parent_type = source_type.GetParentType (context.CurrentThread);
+			source = TryParentCast (context, source, parent_type, target_type);
 			if (source == null)
 				return null;
 
-			return source.GetParentObject (target);
+			return source.GetParentObject (context.CurrentThread);
 		}
 
-		static TargetClassObject TryCurrentCast (ScriptingContext context, Thread target,
+		static TargetClassObject TryCurrentCast (ScriptingContext context,
 							 TargetClassObject source,
 							 TargetClassType target_type)
 		{
-			TargetClassObject current = source.GetCurrentObject (target);
+			TargetClassObject current = source.GetCurrentObject (context.CurrentThread);
 			if (current == null)
 				return null;
 
-			return TryParentCast (context, target, current, current.Type, target_type);
+			return TryParentCast (context, current, current.Type, target_type);
 		}
 
 		public static TargetObject TryCast (ScriptingContext context, TargetObject source,
@@ -2350,23 +2389,21 @@ namespace Mono.Debugger.Frontend
 			if (source.Type == target_type)
 				return source;
 
-			Thread target = context.CurrentThread;
-
-			TargetClassObject sobj = Convert.ToClassObject (target, source);
+			TargetClassObject sobj = Convert.ToClassObject (context.CurrentThread, source);
 			if (sobj == null)
 				return null;
 
 			TargetClassObject result;
-			result = TryParentCast (context, target, sobj, sobj.Type, target_type);
+			result = TryParentCast (context, sobj, sobj.Type, target_type);
 			if (result != null)
 				return result;
 
-			return TryCurrentCast (context, target, sobj, target_type);
+			return TryCurrentCast (context, sobj, target_type);
 		}
 
 		static bool TryParentCast (ScriptingContext context,
-					   TargetClassType source_type,
-					   TargetClassType target_type)
+					   TargetStructType source_type,
+					   TargetStructType target_type)
 		{
 			if (source_type == target_type)
 				return true;
@@ -2374,7 +2411,8 @@ namespace Mono.Debugger.Frontend
 			if (!source_type.HasParent)
 				return false;
 
-			return TryParentCast (context, source_type.ParentType, target_type);
+			TargetStructType parent_type = source_type.GetParentType (context.CurrentThread);
+			return TryParentCast (context, parent_type, target_type);
 		}
 
 		public static bool TryCast (ScriptingContext context, TargetType source,
@@ -2401,13 +2439,13 @@ namespace Mono.Debugger.Frontend
 				if (pexpr != null)
 					address = pexpr.EvaluateAddress (context);
 				else {
-					TargetPointerType ptype = expr.EvaluateType (context)
-						as TargetPointerType;
+					TargetType etype = expr.EvaluateType (context);
+					TargetPointerType ptype = etype as TargetPointerType;
 
 					if ((ptype == null) || ptype.IsTypesafe)
 						throw new ScriptingException (
 							"Cannot cast from {0} to {1}.",
-							ptype.Name, target.Name);
+							etype.Name, target.Name);
 
 					pexpr = new AddressOfExpression (expr);
 					pexpr.Resolve (context);
@@ -2421,6 +2459,7 @@ namespace Mono.Debugger.Frontend
 			TargetClassType type = Convert.ToClassType (target_type);
 			TargetClassObject source = Convert.ToClassObject (
 				context.CurrentThread, expr.EvaluateObject (context));
+
 			if (source == null)
 				throw new ScriptingException (
 					"Variable {0} is not a class type.", expr.Name);
@@ -2444,6 +2483,7 @@ namespace Mono.Debugger.Frontend
 		}
 	}
 
+
 	public static class Convert
 	{
 		static bool ImplicitFundamentalConversionExists (FundamentalKind skind,
@@ -2595,8 +2635,8 @@ namespace Mono.Debugger.Frontend
 		}
 
 		static bool ImplicitReferenceConversionExists (ScriptingContext context,
-							       TargetClassType source,
-							       TargetClassType target)
+							       TargetStructType source,
+							       TargetStructType target)
 		{
 			if (source == target)
 				return true;
@@ -2604,8 +2644,8 @@ namespace Mono.Debugger.Frontend
 			if (!source.HasParent)
 				return false;
 
-			return ImplicitReferenceConversionExists (
-				context, source.ParentType, target);
+			TargetStructType parent_type = source.GetParentType (context.CurrentThread);
+			return ImplicitReferenceConversionExists (context, parent_type, target);
 		}
 
 		static TargetObject ImplicitReferenceConversion (ScriptingContext context,
