@@ -12,20 +12,15 @@ namespace Mono.Debugger.Languages
 			get;
 		}
 
-		public abstract TargetFieldInfo Value {
+		public abstract TargetEnumInfo Value {
 			get;
 		}
 
-		public abstract TargetFieldInfo[] Members {
+		public abstract TargetEnumInfo[] Members {
 			get;
 		}
 
-		internal TargetObject GetValue (TargetLocation location)
-		{
-			return Value.Type.GetObject (location);
-		}
-
-		internal override TargetObject GetObject (TargetLocation location)
+		protected override TargetObject DoGetObject (TargetMemoryAccess target, TargetLocation location)
 		{
 			return new TargetEnumObject (this, location);
 		}
