@@ -1,11 +1,44 @@
+using System;
+using Mono.Debugger.Backend;
+
 namespace Mono.Debugger.Languages
 {
+	internal delegate void MethodLoadedHandler (TargetAccess target, Method method);
+
 	public abstract class TargetFunctionType : TargetType
 	{
 		protected TargetFunctionType (Language language)
 			: base (language, TargetObjectKind.Function)
 		{ }
 
+		public abstract string FullName {
+			get;
+		}
+
+		public abstract bool HasSourceCode {
+			get;
+		}
+
+		public abstract SourceFile SourceFile {
+			get;
+		}
+
+		public abstract int StartRow {
+			get;
+		}
+
+		public abstract int EndRow {
+			get;
+		}
+
+		public abstract bool IsStatic {
+			get;
+		}
+
+		public abstract bool IsConstructor {
+			get;
+		}
+
 		public abstract bool HasReturnValue {
 			get;
 		}
@@ -18,10 +51,6 @@ namespace Mono.Debugger.Languages
 			get;
 		}
 
-		public abstract SourceMethod Source {
-			get;
-		}
-
 		// <summary>
 		//   The current programming language's native representation of
 		//   a method.
@@ -38,10 +67,13 @@ namespace Mono.Debugger.Languages
 			get;
 		}
 
-		public abstract bool IsLoaded {
+		public abstract bool IsManaged {
 			get;
 		}
 
-		public abstract TargetAddress GetMethodAddress (Thread target);
+		internal abstract bool InsertBreakpoint (Thread target,
+							 FunctionBreakpointHandle handle);
+
+		internal abstract void RemoveBreakpoint (Thread target);
 	}
 }
