@@ -1,8 +1,10 @@
 using System;
 
+using Mono.Debugger.Backend;
+
 namespace Mono.Debugger.Languages
 {
-	public abstract class TargetObject : MarshalByRefObject
+	public abstract class TargetObject : DebuggerMarshalByRefObject
 	{
 		internal readonly TargetLocation Location;
 		protected TargetType type;
@@ -38,33 +40,42 @@ namespace Mono.Debugger.Languages
 			}
 		}
 
-		public bool IsNull {
-			get {
-				if (!Location.HasAddress)
-					return false;
-				else
-					return Location.Address.IsNull;
-			}
-		}
-
 		public bool HasAddress {
 			get { return Location.HasAddress; }
 		}
 
-		public TargetAddress Address {
-			get {
-				if (!Location.HasAddress)
-					throw new InvalidOperationException ();
+		public TargetAddress GetAddress (Thread thread)
+		{
+			if (!Location.HasAddress)
+				throw new InvalidOperationException ();
 
-				return Location.Address;
-			}
+			return (TargetAddress) thread.ThreadServant.DoTargetAccess (
+				delegate (TargetMemoryAccess target) {
+					return Location.GetAddress (target);
+			});
 		}
 
-		internal abstract long GetDynamicSize (Thread target, TargetBlob blob,
+		internal TargetAddress GetAddress (TargetMemoryAccess target)
+		{
+			if (!Location.HasAddress)
+				throw new InvalidOperationException ();
+
+			return Location.GetAddress (target);
+		}
+
+		internal abstract long GetDynamicSize (TargetMemoryAccess target, TargetBlob blob,
 						       TargetLocation location,
 						       out TargetLocation dynamic_location);
 
-		public virtual string Print (Thread target)
+		public string Print (Thread thread)
+		{
+			return (string) thread.ThreadServant.DoTargetAccess (
+				delegate (TargetMemoryAccess memory) {
+					return Print (memory);
+			});
+		}
+
+		internal virtual string Print (TargetMemoryAccess memory)
 		{
 			return ToString ();
 		}
