@@ -12,10 +12,10 @@ namespace Mono.Debugger.Languages.Native
 			bounds [0] = new ArrayBounds (lower_bound, upper_bound - lower_bound);
 		}
 
-		protected override void DoGetArrayBounds (Thread target)
+		protected override void DoGetArrayBounds (TargetMemoryAccess target)
 		{ }
 
-		public override TargetObject GetElement (Thread target, int[] indices)
+		internal override TargetObject GetElement (TargetMemoryAccess target, int[] indices)
 		{
 			if (indices.Length != 1)
 				throw new ArgumentException ();
@@ -25,30 +25,39 @@ namespace Mono.Debugger.Languages.Native
 
 			TargetLocation new_location = Location.GetLocationAtOffset (index * size);
 			if (Type.ElementType.IsByRef)
-				new_location = new_location.GetDereferencedLocation (target);
+				new_location = new_location.GetDereferencedLocation ();
 
-			return Type.ElementType.GetObject (new_location);
+			return Type.ElementType.GetObject (target, new_location);
 		}
 
-		public override void SetElement (Thread target, int[] indices,
-						 TargetObject obj)
+		internal override void SetElement (TargetMemoryAccess target, int[] indices,
+						   TargetObject obj)
 		{
 			throw new NotSupportedException ();
 		}
 
-		internal override long GetDynamicSize (Thread target, TargetBlob blob,
+		internal override long GetDynamicSize (TargetMemoryAccess target, TargetBlob blob,
 						       TargetLocation location,
 						       out TargetLocation dynamic_location)
 		{
 			throw new InvalidOperationException ();
 		}
 
-		public override string Print (Thread target)
+		internal override string Print (TargetMemoryAccess target)
 		{
 			if (Location.HasAddress)
-				return String.Format ("{0}", Location.Address);
+				return String.Format ("{0}", Location.GetAddress (target));
 			else
 				return String.Format ("{0}", Location);
 		}
+
+		public override bool HasClassObject {
+			get { return false; }
+		}
+
+		internal override TargetClassObject GetClassObject (TargetMemoryAccess target)
+		{
+			throw new InvalidOperationException ();
+		}
 	}
 }
