@@ -17,6 +17,14 @@ namespace Mono.Debugger.Languages.Native
 
 		TargetType target_type;
 
+		public override bool HasClassType {
+			get { return false; }
+		}
+
+		public override TargetClassType ClassType {
+			get { throw new InvalidOperationException (); }
+		}
+
 		public override bool IsTypesafe {
 			get { return false; }
 		}
@@ -38,7 +46,7 @@ namespace Mono.Debugger.Languages.Native
 			}
 		}
 
-		internal override TargetObject GetObject (TargetLocation location)
+		protected override TargetObject DoGetObject (TargetMemoryAccess target, TargetLocation location)
 		{
 			return new NativePointerObject (this, location);
 		}
